'''ArchitecturalDesigner05.py
  This is a component state-space for PRIMEDesigner05.py.
  Each state component is a FloorPlan object.

  Last updated 31 March 2009.  S. Tanimoto
  Added logging.
  
  DONE: Fix problem with access to s.data after COStar fixup of ops.

  DONE: Fix placement of display when view's scale factor is reduced.

  Changing wall representation: instead of an int, now it is
  an object consisting of two ints and two strings: wallpaper name,
  and wallpaper caption.


  DONE: Provide an operator that deletes all features of a wall,
  including any door.

  DONE: Provision for editing existing caption text, if Enter caption
        is selected on a wall that already has a caption.
        
  Basic functionality is here.
  Things to add:
    1. XML conversion method for the layout.
    2. Detailed display, showing full filenames and caption text.
    3. Smarter shortening of filenames and captions, depending on:
        a. zoom factor, and b. whether on a N or S wall or an E or W wall.
    
'''

import TStar09 as TStar
import COStar09 as COStar
from Tkinter import Canvas

group_name = "Architectural Designer"

class ArchState(TStar.State):
  pass
'''
  def paint_content_in_rectangle(self, a_canvas, xc, yc, w, h):
      new_items = []
      return new_items
'''

class wall:
  def __init__(self):
    self.code = 0
    self.wallpaper = ""
    self.caption = ""
    self.door = -1
    self.puzzle = -1
  def clone(self):
    new_wall = wall()
    new_wall.code = self.code
    new_wall.wallpaper = self.wallpaper
    new_wall.caption = self.caption
    new_wall.door = self.door
    new_wall.puzzle = self.puzzle
    return new_wall
    
def empty_room(): return [wall(),wall(),wall(),wall()]
def row(): return [empty_room(),empty_room(),empty_room()]

class FloorPlan:
  def __init__(self):
    self.rooms = [row(),row(),row()]
    self.room_names = 9*[""]
    self.background_audio_filenames = 9*[""]
    self.room_default_wallpaper_filenames = 9*[""]
    self.overall_default_wallpaper = ""
    self.current_room = [0,0]
    self.facing_dir = 1 # initially facing East.
  def clone(self):
    newfp = FloorPlan()
    r = self.rooms
    for i in range(3):
      for j in range(3):
         for k in range(4):
           newfp.rooms[i][j][k] = (r[i][j][k]).clone()
    newfp.room_names = self.room_names # don't copy by default.
    newfp.background_audio_filenames = self.background_audio_filenames
    newfp.room_default_wallpaper_filenames = self.room_default_wallpaper_filenames
    newfp.current_room = self.current_room[:]
    newfp.facing_dir = self.facing_dir
    newfp.overall_default_wallpaper = self.overall_default_wallpaper
    return newfp
    
  def toStr(self):
    return str(self.rooms)
  def describe(self):
    desc = "Floorplan: \n"
    if self.overall_default_wallpaper:
      desc += " overall_default_wallpaper="+\
              self.overall_default_wallpaper+"\n"
    i = 0
    for row in self.rooms:
     for room in row:
      desc += " Room "+str(i+1)+": "+self.room_names[i]+"\n"
      if self.background_audio_filenames[i]:
        desc += "  bg_audio: "+self.background_audio_filenames[i]+"\n"
      if self.room_default_wallpaper_filenames[i]:
        desc += "  default_wallpaper: "+self.room_default_wallpaper_filenames[i]+"\n"
      i += 1
      for wall in room:
        if wall.code==0 and wall.puzzle==-1 and not wall.wallpaper: continue
        desc += "  wall code="+str(wall.code)+"; "
        if wall.puzzle: desc += "puzzle = "+str(wall.puzzle)+"; "
        if wall.wallpaper:
          desc += "wallpaper = "+wall.wallpaper+"; "
          if wall.caption: desc += "caption = "+wall.caption+"; "
        desc += "\n"
    return desc
    #return self.toStr() # quick and dirty for now.

STARTING_FLOORPLAN = FloorPlan()

#print "STARTING_FLOORPLAN is: " + STARTING_FLOORPLAN.toStr()

def get_dx(dir):
  dx = 1
  if dir==0: return 0
  if dir==2: return 0
  if dir==1: return dx
  if dir==3: return -dx
  return 400

def get_dy(dir):
  dy = 1
  if dir==0: return -dy
  if dir==2: return dy
  if dir==1: return 0
  if dir==3: return 0
  return 400

def shorten_filename(fn):
  n = len(fn); i = n - 1
  while i>0 and fn[i]!='/':
    i -= 1
  ans = fn[i+1:-4]
  return ans

def shorten_caption(cap):
  n = len(cap)
  m = min(n, 10)
  return cap[0:m]

def wall_desc(a_wall):
  c = a_wall.code
  desc = ""
  if c&8>0:
    desc+="d" # closed door
  elif c&1>0:
    desc+="D" # open door
  if a_wall.door and a_wall.door > 0:
    desc+=":"+str(a_wall.door)
    desc+=" "
  if c&2>0:
    if a_wall.puzzle:
      n = int(a_wall.puzzle)
      if n>=100:
        desc+="M"
        n = n-100
      else:
        desc+="P"
      desc+=":"+str(n)
    desc+=" "
  if c&4>0: desc+="W"
  if a_wall.wallpaper:
    shortname = shorten_filename(a_wall.wallpaper)
    desc+="\n"+shortname
  if a_wall.caption:
    shortname = shorten_caption(a_wall.caption)
    desc+="\n"+shortname
  return desc

def has_wallpaper(a_wall):
  return a_wall.code & 4 > 0

def has_no_puzzle(a_wall):
  return  a_wall.code & 2 == 0

def has_no_door_or_puzzle(a_wall):
  return  a_wall.code & 3 == 0

def has_no_wallpaper_or_puzzle(a_wall):
  return a_wall.code & 6 == 0

# A global variable used to represent the floorPlan setup of the initial state:

class ArchNode(TStar.Node):
    def paint_content_in_rectangle(self, a_canvas, xc, yc, w, h):
        new_items = self.s.paint_content_in_rectangle(a_canvas, xc, yc, w, h)
        try: xx = self.p
        except: self.p = []
        self.p += new_items
      
    def domain_specific_init(self, state):
        self.display_elements = []
        self.width = 200
        self.height = 200
        self.rw = self.width/3.0

    def paint_content_in_rectangle(self,a_canvas,x,y,w,h):
        rooms = self.s.data.rooms
        rw = w/3
        half = rw/2
        for i in range(3):
            for j in range(3):
                self.paint_room_in_rectangle(a_canvas, x-rw+j*rw, y-rw+i*rw, rw, rw,
                               rooms[i][j], 3*i+j)
        # Display the arrow showing current room and facing direction:
        z = TStar.ZOOM_FACTOR
        cx = x + (self.s.data.current_room[0]-0.5)*rw-half
        cy = y + (self.s.data.current_room[1]-0.5)*rw-half
        delta = z*10
        dir = self.s.data.facing_dir
        if dir==0: arrow = a_canvas.create_line(cx, cy + delta, cx, cy - delta, arrow="last")
        if dir==1: arrow = a_canvas.create_line(cx - delta, cy, cx + delta, cy, arrow="last")
        if dir==2: arrow = a_canvas.create_line(cx, cy + delta, cx, cy - delta, arrow="first")
        if dir==3: arrow = a_canvas.create_line(cx - delta, cy, cx + delta, cy, arrow="first")
        self.display_elements.append(arrow)
        if (len(self.s.data.overall_default_wallpaper)>0):
          self.display_elements.append(
              a_canvas.create_text(x,y+w/2, text=self.s.data.overall_default_wallpaper, fill="red"))
          

    def paint_room_in_rectangle(self, a_canvas, x, y, w, h, room, room_idx):
        #rw = w #self.rw # room width
        z = TStar.ZOOM_FACTOR
        w2 = w/2
        w5 = w*0.2
        self.display_elements.append(
            a_canvas.create_rectangle(x-w5, y-w5, x+w5, y+w5, fill="red"))

        x0 = x-w2; y0 = y-w2; x1=x0; x2 = x0+w*0.3; x3=x2; y1 = y0+w;
        y2 = y-w2+w*0.7 ; y3 = y-w2+w*0.3
        self.display_elements.append(
            a_canvas.create_polygon(x0,y0,x1,y1,x2,y2,x3,y3, fill="blue", outline="black"))

        x0 = x-w2; y0 = y-w2+w; x1=x0+w; x2 = x0+w*0.7; x3=x0+w*0.3;
        y1 = y-w2+w; y2 = y-w2+w*0.7; y3 = y2
        self.display_elements.append(
            a_canvas.create_polygon(x0,y0,x1,y1,x2,y2,x3,y3, fill="green", outline="black"))

        x0 = x-w2+w; y0 = y-w2+w; x1=x0; y1=y-w2; x2 = x-w2+w*0.7; y2 = y-w2+w*0.3;
        x3=x2; y3 = y-w2+w*0.7;
        self.display_elements.append(
            a_canvas.create_polygon(x0,y0,x1,y1,x2,y2,x3,y3, fill="yellow", outline="black"))

        x0 = x-w2; y0 = y-w2; x1=x0+w; y1=y0; x2 = x0+w*0.7; y2 = y0+w*0.3;
        x3=x0+w*0.3; y3 = y2;
        self.display_elements.append(
            a_canvas.create_polygon(x0,y0,x1,y1,x2,y2,x3,y3, fill="orange", outline="black"))

        # If the room has a name, display it just below the center of the room:
        names = self.s.data.room_names
        name = names[room_idx]
        if len(name)>0:
          self.display_elements.append(
              a_canvas.create_text(x,y-w/8, text=name, fill="white"))
        bg_audio_filenames = self.s.data.background_audio_filenames
        bga_filename = bg_audio_filenames[room_idx]
        if len(bga_filename)>0:
          if TStar.ZOOM_FACTOR > 0.9:
            if len(bga_filename)>22: bga_filename = bga_filename[-20:]
          else:
            if len(bga_filename)>12: bga_filename = bga_filename[-12:]
          self.display_elements.append(
              a_canvas.create_text(x,y+w/8, text=bga_filename, fill="white"))
        default_wallpaper_list = self.s.data.room_default_wallpaper_filenames
        default_wallpaper = default_wallpaper_list[room_idx]
        if len(default_wallpaper)>0:
          if len(default_wallpaper)>12: default_wallpaper = default_wallpaper[-12:]
          self.display_elements.append(
              a_canvas.create_text(x,y+w/4, text=default_wallpaper, fill="white"))
         
        # Paint the features of the walls in the room
        #xr = x + rw/2; yr = y + rw/2
        xr = x; yr = y
        #print "Painting room"+str(room)
        for dir in range(4):
          wall = room[dir]
          q = 0.37*w
          dx = get_dx(dir)*q
          dy = get_dy(dir)*q
          self.display_elements.append(
            a_canvas.create_text(xr+dx, yr+dy,
                                 text=wall_desc(wall),
                                 justify="center"))

    def unpaint_content(self,a_canvas):
        """deletes the node's blocks and labels from the canvas."""
        if self.has_painted_content:
            for elt in self.display_elements:
                a_canvas.delete(elt)
            self.display_elements = []
        self.has_painted_content=False

'''Now we define the operators, beginning with some helping functions.'''

def face(fp, dir):
    newfp = fp.clone()
    newfp.facing_dir = dir
    return newfp

def ok_to_move(fp):
    d = fp.facing_dir
    if d==0: return fp.current_room[1]>0
    if d==1: return fp.current_room[0]<2
    if d==2: return fp.current_room[1]<2
    if d==3: return fp.current_room[0]>0
    return False

def move_to_next_room(fp):
    newfp = fp.clone()
    d = fp.facing_dir
    if d==0: newfp.current_room[1]-=1
    if d==1: newfp.current_room[0]+=1
    if d==2: newfp.current_room[1]+=1
    if d==3: newfp.current_room[0]-=1
    return newfp

def wall_is_plain(fp):
    j = fp.current_room[0]
    i = fp.current_room[1]
    r = fp.rooms[i][j]
    w = r[fp.facing_dir]
    if w.code==0: return True
    return False

def get_room_name(fp):
    k = fp.current_room[1]*3+fp.current_room[0]
    return fp.room_names[k]

def set_room_name(fp, name):
    k = fp.current_room[1]*3+fp.current_room[0]
    newfp = fp.clone()
    newfp.room_names = fp.room_names[:] # Copy
    newfp.room_names[k] = name
    return newfp

def set_background_music(fp,filename): 
    k = fp.current_room[1]*3+fp.current_room[0]
    newfp = fp.clone()
    newfp.background_audio_filenames = fp.background_audio_filenames[:] # Copy
    newfp.background_audio_filenames[k] = filename
    return newfp
  
def configure_wall(fp, wall_feature, wallpaper=None, caption=None,
                   door_num=None, puzzle_num=None):
    newfp = fp.clone()
    j = fp.current_room[0]
    i = fp.current_room[1]
    d = fp.facing_dir
    #print "configure_wall: i,j are: "+ str(i)+", "+str(j)
    newfp.rooms[i][j][d].code |= wall_feature  # inclusive OR the bit into the desc.
    # If wallpaper is specified, use this string instead of the number.
    if wallpaper: (newfp.rooms[i][j][d]).wallpaper = wallpaper
    if caption: (newfp.rooms[i][j][d]).caption = caption
    if door_num: newfp.rooms[i][j][d].door = door_num
    if puzzle_num: newfp.rooms[i][j][d].puzzle = puzzle_num
    # If placing a door, then replicate door on opposite side of wall
    # If no door_num is provided, make up one by concatenating the
    # room numbers of the rooms being joined, taking the smaller first.
    if wall_feature == 1 or wall_feature == 8:
      # Create open door or close the open door.
      next_room = get_next_room(newfp, i, j, d)
      next_room[(d+2)%4].code |= wall_feature
      if wall_feature==1:
        newfp.rooms[i][j][d].code &= 7  # open the door if it is closed
        next_room[(d+2)%4].code   &= 7  
      if not door_num:
        n1 = 3*i+j+1
        n2 = get_next_room_index(newfp, i, j, d)+1
        if n1<=n2: door_num = 10*n1+n2
        else: door_num = 10*n2+n1
        newfp.rooms[i][j][d].door = door_num
        next_room[(d+2)%4].door = door_num
    return newfp

def set_default_room_wallp(fp, wallpaper=None):
    k = fp.current_room[1]*3+fp.current_room[0]
    newfp = fp.clone()
    newfp.room_default_wallpaper_filenames = fp.room_default_wallpaper_filenames[:] # Copy
    newfp.room_default_wallpaper_filenames[k] = wallpaper
    return newfp

def set_default_wallp(fp, wallpaper=None):
    newfp = fp.clone()
    newfp.overall_default_wallpaper = wallpaper
    return newfp

def clear_wall(fp):
    # Get rid of any door, wallpaper, caption, and/or puzzle.
    # This convenience operator can help designers reduce
    # the amount of backtracking they have to do to revise
    # a design decision.
    w = current_wall(fp)
    newfp = fp.clone()
    w = current_wall(newfp)
    if w.code&1 > 0: # Need to remove door from adjacent room
      j = fp.current_room[0]
      i = fp.current_room[1]
      d = fp.facing_dir
      next_room = get_next_room(newfp, i, j, d)
      next_room[(d+2)%4].code &= 6 # Turn off bit 0
      next_room[(d+2)%4].door = -1
    w.code = 0
    w.door = -1
    w.puzzle = -1
    w.wallpaper = None
    w.caption = None
    return newfp
  
def current_wall(fp):
    j = fp.current_room[0]
    i = fp.current_room[1]
    r = fp.rooms[i][j]
    d = fp.facing_dir
    w = r[d]
    return w
  
def ok_for_door(fp):
    j = fp.current_room[0]
    i = fp.current_room[1]
    r = fp.rooms[i][j]
    d = fp.facing_dir
    w = r[d]
    if not has_no_door_or_puzzle(w): return False
    if d==0:
      if i==0: return False
    if d==1:
      if j==2: return False
    if d==2:
      if i==2: return False
    if d==3:
      if j==0: return False
    next_room = get_next_room(fp, i, j, d)
    w2 = next_room[(d+2)%4]
    if not has_no_door_or_puzzle(w2): return False
    return True

def ok_for_puzzle(fp):
    return wall_is_plain(fp)

def ok_for_wallpaper(fp):
    return has_no_wallpaper_or_puzzle(current_wall(fp))

def ok_for_selection(fp):
    return has_no_puzzle(current_wall(fp))
  
def has_wallpaper(fp):
    return current_wall(fp).code & 4 > 0

def door_is_closed(fp):
    return current_wall(fp).code & 8 > 0
  
def door_is_open(fp):
    return current_wall(fp).code & 9 == 1 # door exists and is not closed.

def get_next_room(fp, i, j, d):
    ni = i; nj = j
    if d==0: ni = i-1
    if d==1: nj = j+1
    if d==2: ni = i+1
    if d==3: nj = j-1
    return fp.rooms[ni][nj]

def get_next_room_index(fp, i, j, d):
    ni = i; nj = j
    if d==0: ni = i-1
    if d==1: nj = j+1
    if d==2: ni = i+1
    if d==3: nj = j-1
    return ni*3+nj

def dirNot(dat, dir):
    d = dat.facing_dir
    return d!= dir

'''
def get_caption(s):
    # This function will be used for initializing the dialogs for
    # editing wallpaper captions.
    # The function is designed directly for COStar and does not
    # need to be fixed up.  But that means this will not work
    # without COStar.
    fp  = s.data[0]
    print "in get_caption, fp is "+str(fp)
    w = current_wall(fp)
    return w.caption
'''
def get_caption(fp):
    print "in get_caption, fp is "+str(fp)
    return current_wall(fp).caption
  
def initializer():
    # Returns valid initial data for this kind of state
    global STARTING_FLOORPLAN
    return STARTING_FLOORPLAN

def setup():    
    Face_north = TStar.Operator("Face north",
               lambda b,p: face(b,0),
               precondition=lambda s: dirNot(s,0),
               group=group_name)
    Face_east = TStar.Operator("Face east",
               lambda b,p: face(b,1),
               precondition=lambda s: dirNot(s,1),
               group=group_name)
    Face_west = TStar.Operator("Face west",
               lambda b,p: face(b, 3),
               precondition=lambda s: dirNot(s,3),
               group=group_name)
    Face_south = TStar.Operator("Face south",
               lambda b,p: face(b,2),
               precondition=lambda s: dirNot(s,2),
               group=group_name)
    Next_room = TStar.Operator("Move to next room",
               lambda b,p: move_to_next_room(b),
               precondition=lambda s: ok_to_move(s),
               group=group_name)
    Name_room = TStar.OperatorSpecifiedWithDialog("Edit name of room",
               lambda b,text,p: set_room_name(b,text),
               group=group_name,
               prompt="Enter your name for this room...",
               initial_value_fn=get_room_name)
    Choose_background_music = TStar.OperatorSpecifiedWithDialog("Choose background music",
               lambda b,filename,p: set_background_music(b,filename),
               group=group_name,
               param_op_type='File',
               prompt="Select an audio file for this room's background music...",
               filter_desc=[('MP3 Files','*.MP3'),('WAV Files','*.WAV')])
    Place_door = TStar.Operator("Place door",
               lambda b,p: configure_wall(b,1),
               precondition=lambda s: ok_for_door(s),
               group=group_name)
    Close_door = TStar.Operator("Close door",
               lambda b,p: configure_wall(b,8),
               precondition=lambda s: door_is_open(s),
               group=group_name)
    Open_door = TStar.Operator("Open door",
               lambda b,p: configure_wall(b,1),
               precondition=lambda s: door_is_closed(s),
               group=group_name)
#    Place_puzzle = TStar.Operator("Place puzzle",
#               lambda b,p: configure_wall(b,2),
#               precondition=lambda s: ok_for_puzzle(s),
#               group=group_name)
    Place_specific_image_puzzle = TStar.OperatorSpecifiedWithDialog("Place Image Puzzle with Number",
               lambda b,num,p: configure_wall(b,2,puzzle_num=num),
               precondition=lambda s: ok_for_puzzle(s),
               group=group_name,
               param_op_type='Int',
               prompt="Enter an integer to control what puzzle goes here...")
    Place_specific_music_puzzle = TStar.OperatorSpecifiedWithDialog("Place Music Puzzle with Number",
               lambda b,num,p: configure_wall(b,2,puzzle_num=str(int(num)+100)),
               precondition=lambda s: ok_for_puzzle(s),
               group=group_name,
               param_op_type='Int',
               prompt="Enter an integer to control what puzzle goes here...")
    Place_wallpaper = TStar.Operator("Indicate use of wallpaper",
               lambda b,p: configure_wall(b,4),
               precondition=lambda s: ok_for_wallpaper(s),
               group=group_name)
    Place_specific_wallpaper = TStar.OperatorSpecifiedWithDialog("Select wallpaper",
               lambda b,filename,p: configure_wall(b,4,wallpaper=filename),
               precondition=lambda s: ok_for_selection(s),
               group=group_name,
               param_op_type='File',
               prompt="Select an image for this wall's wallpaper...",
               filter_desc=[('JPG Files','*.JPG'),('GIF Files','*.GIF'),
                            ('PNG Files','*.PNG'),('BMP Files','*.BMP')])
    Set_game_wallpaper = TStar.OperatorSpecifiedWithDialog("Set default wallpaper for entire game",
               lambda b,filename,p: set_default_wallp(b,wallpaper=filename),
               precondition=lambda s: True,
               group=group_name,
               param_op_type='File',
               prompt="Select an image for the game's default wallpaper...",
               filter_desc=[('JPG Files','*.JPG'),('GIF Files','*.GIF'),
                            ('PNG Files','*.PNG'),('BMP Files','*.BMP')])
    Set_room_wallpaper = TStar.OperatorSpecifiedWithDialog("Set room's default wallpaper",
               lambda b,filename,p: set_default_room_wallp(b,wallpaper=filename),
               precondition=lambda s: True,
               group=group_name,
               param_op_type='File',
               prompt="Select an image for this room's default wallpaper...",
               filter_desc=[('JPG Files','*.JPG'),('GIF Files','*.GIF'),
                            ('PNG Files','*.PNG'),('BMP Files','*.BMP')])
    Enter_caption = TStar.OperatorSpecifiedWithDialog("Provide caption",
               lambda b,text,p: configure_wall(b,4,caption=text),
               precondition=lambda s: has_wallpaper(s),
               group=group_name,
               #param_op_type='Free Text',
               prompt="Enter a description for the wallpaper image...",
               initial_value_fn=get_caption)
    Clear_wall = TStar.Operator("Remove all wall features",
               lambda b,p: clear_wall(b),
               precondition=lambda s: not wall_is_plain(s),
               group=group_name)

    ArchOps = [Face_north, Face_east, Face_west, Face_south,
          Next_room, Name_room, Choose_background_music, Place_door, Close_door,
               Open_door, Place_specific_image_puzzle,
               Place_specific_music_puzzle,
               Place_specific_wallpaper, Set_game_wallpaper,
               Set_room_wallpaper, Place_wallpaper,
               Enter_caption, Clear_wall]

    COStar.register_state_space_component(group_name,ArchState, ArchNode,
                                      ArchOps,
                                      initializer)






