''' GraphicGameTester.py
  Written by S. Tanimoto as part of the NSF-funded project on
  "Problem-solving methodology in collaborative design."

  (C) 2007 by the University of Washington and Steven Tanimoto.

   
This file adapts GameTester.py  to use a GUI with TKinter

There were 5 phases of work to do in developing this program:

1. Create a window to hold all the panes. (DONE)

2. Create a text area to host the basic interaction of GameTester. (DONE)
      The game engine telling the user where s/he is, etc.
      Textual input from the user.
      
3. Create a Current Wall view panel, and implement the following:
     (a) Display of wallpaper on that wall. (DONE)
     (b) Display of a door on that wall, with open and closed states. (DONE) 
     (c)   When open, maybe show any wall paper on the opposite wall of the
           next room. -- NO, too messy, esp. if several doors are open in a line.
     (d) Display puzzle information for Image puzzles.
         Puzzle IMAGES now appear. (DONE)
     (e) Display music information for music puzzles. (WORKING)
     (f) Play music associated with music puzzles. (WORKING)
     (g) Display of floor, ceiling, wall to left and wall to right, as trapezoids for
        spatial context. (NOT DONE)

4. Create a Maze Map view, to show the layout of the game, the current location
      of the player, and the states of the doors (open or closed).
      Maybe also the states of the puzzles (unattempted, attempted, solved).
      (STARTED)
      
5. Render music as music in music puzzles.
      Include a music on/off button on the GUI.  
      There can be a switch on the GUI to turn off automatically playing the
         current melody when entering a room or facing the wall with the puzzle.
         If playing is on by default and there are 2 or more music puzzles in
         the room, do they both play simultaneously?

6. Create Score panel.

Status:
   STARTED on 16 June 2007
   Steps 1 and 2 completed on 3 August 2007.
   Step 3 in progress on 4 August, with (a), (b), and (d) completed Aug 8.
   All the functionality of the textual engine in GameTester.py now
    works in this version using the TEXT_OUTPUT_PANE and INPUT_PANE
    areas in the new window.

   Maze Map shows the rooms in the design and the current
   room and current facing direction.  Doors, door open/closed states,
   puzzle existence, or puzzle state are shown in a rudimentary fashion.
   Puzzle state is 'p' for unsolved and 'P' for solved.

   21 Aug 2007: Added a Score pane.

   13 Sept 2007: Works when invoked from a PRIME-Designer menu.

'''
import PRIMEClasses         # Representation system for PRIME games.

# GameTester.py contains a basic textual game engine.
import GameTester           # Here we redefine the text output method in GameTester
# as well as a few of its other methods.


# The following are needed in order to display GIF and JPG images:
import Image as PIL
from Tkinter import * # Tkinter is a popular a GUI development library for Python.
from ImageTk import PhotoImage

# The following is needed to play audio files. It requires that
# the FMOD audio DLL file be installed on Windows, and that that
# PySonic module be downloaded and installed.
try:
    from MusicP.PySonicPlayAudio import play_audio_file, pause_audio
    AUDIO_FILE_PLAYING_WORKS = True
except:
    AUDIO_FILE_PLAYING_WORKS = False
    print "Audio file playing won't work due to missing module: PySonicPlayAudio"
    def play_audio_file(x): return None
    def pause_audio(): return None
    

# Here are some global variables used for the GUI:   
TK_ROOT = None; WV_CANVAS = None; MZ_CANVAS = None; RULE_PANE = None
TEXT_OUTPUT_PANE = None; INPUT_PANE = None

# The following global always indicates one of three possible states
# of interaction:
INPUT_ACTIVITY = "navigation" # vs "puzzle_solving" or "uttering_magic"

# A BOOLEAN TO CONTROL MUSIC ON/OFF:
PLAYING_MUSIC = True

def setup_and_run(tkroot=None):
  global TK_ROOT, WV_CANVAS, MZ_CANVAS, TEXT_OUTPUT_PANE, INPUT_PANE
  global RULE_PANE
  # Create the GUI's window:
  if tkroot==None:
    TK_ROOT = Tk(className=" PRIME Game Playing Window")
  else:
    TK_ROOT = Toplevel(tkroot)
  TK_ROOT.geometry('750x600+100+25') # Sets size and position.
  TK_ROOT.grid_rowconfigure(0, weight=1)
  TK_ROOT.grid_columnconfigure(0, weight=1)
  # Create the canvas where views of the current wall will be drawn:
  WV_CANVAS = Canvas(TK_ROOT, bg='blue')
  WV_CANVAS.grid(row=0,column=0,columnspan=2,sticky='news')
  WV_CANVAS.create_text(180,10,text="Wall View Panel")
  # Create the canvas where the maze will be drawn when enabled:
  MZ_CANVAS = Canvas(TK_ROOT, bg='purple')
  MZ_CANVAS.grid(row=0,column=2,sticky='news')
  MZ_CANVAS.create_text(200,10,text="Maze Map Panel")
  # Create a scrollable text area in which the game's text output
  # will be shown to the user:
  scrollbar = Scrollbar(TK_ROOT, orient=VERTICAL)
  scrollbar.grid(row=1,column=1,sticky=N+S) 
  TEXT_OUTPUT_PANE = Text(TK_ROOT, width=80, height=10,yscrollcommand=scrollbar.set)
  scrollbar.config(command=TEXT_OUTPUT_PANE.yview)
  TEXT_OUTPUT_PANE.insert("1.0", "Interaction text area\n")
  TEXT_OUTPUT_PANE.grid(row=1,column=0,sticky='sw')
  # Create a text entry field, where the user will enter commands:
  INPUT_PANE = Entry(TK_ROOT, width=80)
  INPUT_PANE.bind("<Return>",handle_input)
  INPUT_PANE.grid(row=2,column=0,sticky='nw')
  INPUT_PANE.insert(END, "Replace this line of text with your move, and type the Enter key.")
  INPUT_PANE.focus_force()
  INPUT_PANE.select_range(0, END)
  # Create a score pane:
  RULE_PANE = Canvas(TK_ROOT, height=50, bg='gray')
  RULE_PANE.grid(row=1,column=2 ,sticky='news')
  show_score()
  # Now start the game!
  GameTester.do_welcome()
  display_maze()
  show_wall()
  main_prompt()
  TK_ROOT.mainloop()

def show_interaction_text(msg):
    # This appends a message into the scrollable text area.
    global TEXT_OUTPUT_PANE
    TEXT_OUTPUT_PANE.configure(state="normal")
    TEXT_OUTPUT_PANE.insert(END, msg+"\n")
    TEXT_OUTPUT_PANE.see(END)
    TEXT_OUTPUT_PANE.configure(state="disabled")

def handle_input(dummy):
    # This responds to an Enter event in the INPUT_PANE.
    global INPUT_ACTIVITY
    theInput = INPUT_PANE.get()
    TEXT_OUTPUT_PANE.insert(END, theInput+"\n")
    INPUT_PANE.delete(0, END)
    if INPUT_ACTIVITY == "navigation":
        done=GameTester.do_one_iteration(theInput)
        if done: show_interaction_text("""\nTHANKS FOR PLAYING!""")
        else:
            show_wall()
            display_maze()
            if INPUT_ACTIVITY == "navigation": main_prompt()
    elif INPUT_ACTIVITY == "puzzle_solving":
        print "Received input while working on a puzzle."
        handle_puzzle_move(theInput)
        show_wall()
        display_maze()
    elif INPUT_ACTIVITY == "uttering_magic":
        handle_magic_selection(theInput)
    show_score()

def change_room(new_room):
  if AUDIO_FILE_PLAYING_WORKS:
    audio_file = new_room.background_audio
    if (audio_file):
        full_filename = PRIMEClasses.FOLDER_NAME+"/"+audio_file
        play_audio_file(full_filename)
    else: pause_audio()

def main_prompt():
    show_interaction_text('''Make a move (H for help).''')
    show_interaction_text('''---------------------------------------''')

# Replace the output routine in GameTester by the new one:
GameTester.tell_user=show_interaction_text

# Here are some global constants needed for displaying the wall view.
WALL_VIEW_WIDTH = 360; WALL_VIEW_HEIGHT = 360
WALL_WIDTH = 300; WALL_HEIGHT = 300
WALL_X1 = (WALL_VIEW_WIDTH  - WALL_WIDTH )/2; WALL_X2 = WALL_X1 + WALL_WIDTH
WALL_Y1 = (WALL_VIEW_HEIGHT - WALL_HEIGHT)/2; WALL_Y2 = WALL_Y1 + WALL_HEIGHT
DOOR_WIDTH = 80
DOOR_X1 = WALL_VIEW_WIDTH/2 - DOOR_WIDTH/2
DOOR_X2 = WALL_VIEW_WIDTH/2 + DOOR_WIDTH/2
DOOR_HEIGHT = 150
DOOR_Y1 = WALL_VIEW_HEIGHT/2 - DOOR_HEIGHT/2
DOOR_Y2 = WALL_VIEW_HEIGHT/2 + DOOR_HEIGHT/2
PUZZLE_WIDTH = 256; PUZZLE_HEIGHT = 256
PUZZLE_X1 = WALL_VIEW_WIDTH/2  - PUZZLE_WIDTH/2
PUZZLE_Y1 = WALL_VIEW_HEIGHT/2 - PUZZLE_HEIGHT/2
PUZZLE_X2 = WALL_VIEW_WIDTH/2  + PUZZLE_WIDTH/2
PUZZLE_Y2 = WALL_VIEW_HEIGHT/2 + PUZZLE_HEIGHT/2

# We keep track of display items so they can be easily removed when it's
# time to clear the canvas before showing the next view.
WALL_CANVAS_ITEMS = []

def show_wall():
    global PHOTOIMAGES, WV_CANVAS, WALL_CANVAS_ITEMS
    global IMAGE_PUZZLES, MUSIC_PUZZLES, PLAYING_MUSIC
    # Clear display of the previous wall:
    for item in WALL_CANVAS_ITEMS:
        WV_CANVAS.delete(item)
    WALL_CANVAS_ITEMS = []
    # Now display the new wall:
    w = GameTester.get_current_wall()
    #print "In show_wall, current wall is "+w.get_description()
    desc = wall_desc_str(w)
    item = WV_CANVAS.create_text(WALL_VIEW_WIDTH/2,WALL_VIEW_HEIGHT +20, text=desc)
    WALL_CANVAS_ITEMS.append(item)
    if w.wallpaper != None:
        #print "Trying to get and show wallpaper based on file: "+w.wallpaper
        try: 
            get_image(w.wallpaper)
            item = WV_CANVAS.create_image(WALL_X1, WALL_Y1,image=PHOTOIMAGES[w.wallpaper],anchor=NW)
            WALL_CANVAS_ITEMS.append(item)
        except:
            print "In get_image, could not open the file: "+PRIMEClasses.FOLDER_NAME+"/"+w.wallpaper
    if w.text != None:
        item = WV_CANVAS.create_text(WALL_VIEW_WIDTH/2, WALL_Y1+WALL_VIEW_HEIGHT+10,text=w.text,anchor=CENTER,fill="dark red")
        WALL_CANVAS_ITEMS.append(item)
    if w.door != None:
        if w.door.open:
            #item = WV_CANVAS.create_rectangle(DOOR_X1, DOOR_Y1, DOOR_X2, DOOR_Y2,fill="light blue")
            item = WV_CANVAS.create_polygon(DOOR_X1, DOOR_Y1, DOOR_X2, DOOR_Y1, DOOR_X2, DOOR_Y2, DOOR_X1, DOOR_Y2,
                                            outline="green",width=6,fill="")
        else:
            item = WV_CANVAS.create_polygon(DOOR_X1, DOOR_Y1, DOOR_X2, DOOR_Y1, DOOR_X2, DOOR_Y2, DOOR_X1, DOOR_Y2,
                                            outline="red",width=3,fill="")
        WALL_CANVAS_ITEMS.append(item)

    if w.puzzle_num != None:
        thePuzzle = PRIMEClasses.get_puzzle(w.puzzle_num)
        if (thePuzzle==None):
            print "No puzzle number "+str(w.puzzle_num)+" exists."
            return
        # Show puzzle frame:
        item = WV_CANVAS.create_polygon(PUZZLE_X1, PUZZLE_Y1, PUZZLE_X2, PUZZLE_Y1,
                                        PUZZLE_X2, PUZZLE_Y2, PUZZLE_X1, PUZZLE_Y2,
                                        outline="brown",width=10,fill="")
        WALL_CANVAS_ITEMS.append(item)
        # Show puzzle title:
        item = WV_CANVAS.create_text(WALL_VIEW_WIDTH/2,(WALL_VIEW_HEIGHT - PUZZLE_HEIGHT)/2 - 15,
                                     text="Puzzle:" + thePuzzle.name)
        WALL_CANVAS_ITEMS.append(item)
        puzzle_type = thePuzzle.get_puzzle_type()
        print "The type of this puzzle is: "+puzzle_type
        # If the current configuration of the puzzle has an image associated with it, display it.
        #if puzzle_type=="image":
        cos = thePuzzle.current_op_sequence
        # Find matching image:
        if puzzle_type=="image":
          procImages = thePuzzle.proc_images
          try:
            imfn = procImages[tuple(cos)]
          except:
            try:
              imfn = procImages["default"]
            except: imfn = "NO-DEFAULT-FILENAME-GIVEN-IN-PUZZLE"
        else:
          imfn = thePuzzle.background
        print "The file to show is "+imfn
        # Load image
        get_image(imfn, True)
        try:
          if PHOTOIMAGES[imfn]:
            print "In GraphicGameTester.show_wall, got the puzzle image: "+imfn
            # Display image
            try: 
              item = WV_CANVAS.create_image(WALL_VIEW_WIDTH/2,WALL_VIEW_HEIGHT/2,
                            image=PHOTOIMAGES[imfn],anchor=CENTER)
              WALL_CANVAS_ITEMS.append(item)
            except:
              print "Could not create the the canvas image for "+imfn
              raise
          else:
            print "In GraphicGameEngine.py, although image was loaded, it's not avail. Strange."
        except: 
            print "No image available with name: "+imfn
            raise
        if puzzle_type=="music":
            # Display the textual note sequence on the wall.
            try:
                #print "thePuzzle.note_seq="+str(thePuzzle.note_seq)
                #print "current operator sequence: "+str(cos)
                notes = thePuzzle.note_seq[tuple(cos)]
            except:
                notes = "NO MESSAGE PROVIDED"
            item = WV_CANVAS.create_text(WALL_VIEW_WIDTH/2,WALL_VIEW_HEIGHT/4,
                          text="Musical note sequence is: ",anchor=CENTER)
            WALL_CANVAS_ITEMS.append(item)
            item = WV_CANVAS.create_text(WALL_VIEW_WIDTH/2,WALL_VIEW_HEIGHT/2,
                          text=notes,anchor=CENTER, font=30, fill="white")
            WALL_CANVAS_ITEMS.append(item)
            #print "Now we should play the music"
            if PLAYING_MUSIC:
                play_notes(notes)
    return

# Maintain a hash table of images used in wallpaper and in puzzles, so that,
# once loaded and turned into Tkinter "photoimages", they can be re-used.
PHOTOIMAGES = {}
# The following method allows the PRIME_Design program to override it,
# to work around a bug in Tkinter that leads to PhotoImage failing when
# reimported.
def create_photoimage(pil_image):
  return PhotoImage(pil_image)

def get_image(filename, puzzle_image=False):
    global PHOTOIMAGES, PUZZLE_WIDTH, PUZZLE_HEIGHT, WALL_WIDTH, WALL_HEIGHT
    try:
        if PHOTOIMAGES[filename]!=None:
          print "in GraphicGameTester.get_image, PHOTOIMAGES["+filename+"] exists."
          return # Image is already cached.
    except: print "There was a problem looking up PHOTOIMAGES["+filename+"]."
    print "In GraphicGameTester, line 281, PRIMEClasses.FOLDER_NAME is "+PRIMEClasses.FOLDER_NAME
    print "filename is "+filename
    full_filename = PRIMEClasses.FOLDER_NAME+"/"+filename
    print "full_filename = "+full_filename
    try:
        temp_image = PIL.open(full_filename)
        print "In GraphicGameTester.get_image, successfully opened the image file: "+full_filename+"."
    except:
        print "Image file not found: "+full_filename
        return
    if puzzle_image:
      print "This is a puzzle image."
      try:
        temp_image2 = temp_image.resize((PUZZLE_WIDTH-20,PUZZLE_HEIGHT-20))
      except:
        print "Could not resize the puzzle image."
    else:
        temp_image2 = temp_image.resize((WALL_WIDTH,WALL_HEIGHT))
    try:
      PHOTOIMAGES[filename] = create_photoimage(temp_image2)
      return
    except:
      print "Could not make PHOTOIMAGE."
      
# Because interaction is now event-based, the handling of
# uttering magic phrases and solving puzzles is a little
# different from the way it is in GameTester.py.
    
# The following two methods replace the maybe_utter method of
# GameTester to provide event-driven interaction.
def GGT_maybe_utter():
    phrases = GameTester.MAGIC_PHRASE_INVENTORY
    if len(phrases)==0:
        tell_user("You need to know magic to utter magic words.")
        return
    tell_user("You have the following magic phrases available:")
    counter = 0
    for mp in phrases:
        tell_user(str(counter)+': '+phrases[counter])
        counter += 1
    tell_user("Enter the number for the magic phrase you with to say, or -1 to cancel.")
    global INPUT_ACTIVITY
    INPUT_ACTIVITY = "uttering_magic"
    return

def handle_magic_selection(theInput):
    global INPUT_ACTIVITY
    numstring = theInput
    n = -1
    try:
        n = int(numstring)
    except:
        tell_user("Your typed:"+numstring+". That's not a number I recognize. Try again!")
        return
    if n<0:
        tell_user("Maybe you'll be more ready to engage in magic next time.")
        INPUT_ACTIVITY = "navigation"
        return
    if n>=len(GameTester.MAGIC_PHRASE_INVENTORY):
        tell_user("That number is too big. Try again.")
        return
    mp = GameTester.MAGIC_PHRASE_INVENTORY[n]
    num = GameTester.MAGIC_PHRASE_PUZZLE_NUM[n]
    tell_user("You have chosen to utter the phrase:")
    tell_user("   "+mp)
    GameTester.handle_event("Uttering", num)
    tell_user("There MIGHT or MIGHT NOT be magic happening.")
    GameTester.MAGIC_PHRASE_INVENTORY.remove(mp)
    GameTester.MAGIC_PHRASE_PUZZLE_NUM.remove(num)
    INPUT_ACTIVITY = "navigation"
    show_wall()
    display_maze()
    main_prompt()

# Global variables to keep track of what puzzle is being worked on:
CURRENT_PUZZLE = None
PUZZLE_NUMBER = None
# The following method handles the "P" command to start working on
# the puzzle that is presumably on the wall in front of the avatar.
def GGT_do_puzzle_activity():
    ''' Handles puzzle activity by the player.
    '''
    global CURRENT_PUZZLE, PUZZLE_NUMBER
    PUZZLE_NUMBER = GameTester.get_any_puzzle_here()
    if PUZZLE_NUMBER < 0:
        tell_user("There's no puzzle in front of you to work on.")
        return

    CURRENT_PUZZLE = GameTester.get_puzzle(PUZZLE_NUMBER)
    tell_user("You are now working on puzzle "+CURRENT_PUZZLE.name)
    CURRENT_PUZZLE.describe()
    solution = list(eval(CURRENT_PUZZLE.correct_sequence))
    while(True):
        tell_user("The state of your work on this puzzle is represented by ")
        cos = CURRENT_PUZZLE.current_op_sequence
        tell_user("  the operator sequence: " + str(cos))
        if CURRENT_PUZZLE.solved:
            tell_user("You have already solved this puzzle. (Remember?)")
            return
        tell_user("Here are your choices:")
        ops = CURRENT_PUZZLE.operators.values()
        ops.sort()
        n = len(ops)
        for op in ops:
            tell_user("   "+op.to_string())
        tell_user("   -1: undo last operator")
        tell_user("   L:  Leave puzzle in its current state")
        global INPUT_ACTIVITY
        INPUT_ACTIVITY = "puzzle_solving"
        return

# The next method assumes the player is working on a puzzle and
# it processes the current input as a puzzle command.
def handle_puzzle_move(theInput):
    global INPUT_ACTIVITY, CURRENT_PUZZLE, PUZZLE_NUMBER
    c = theInput[:]
    good_choice = False
    if c=='L' or c=='l':
        INPUT_ACTIVITY = "navigation"
        tell_user("You are leaving this puzzle in its current state.")
        main_prompt()
        return
    try:
        k = int(c)
    except: 
        tell_user("Your input: "+c+" is not a valid command here."+\
                  "Type an integer or the letter L.")
        return
    ops = CURRENT_PUZZLE.operators.values()
    ops.sort()
    n = len(ops)
    cos = CURRENT_PUZZLE.current_op_sequence
    if k > -1 and k < n: good_choice = True
    else:
        if k==-1:
            if len(cos)>0: good_choice = True
            else:
                tell_user("There is nothing to undo.")
                return
    if not good_choice:
        tell_user("The choice: "+c+" is invalid. Try again.")
        return
    
    if k==-1:
        cos.pop()
        tell_user("Undoing last operator.")
        return
    cos.append(k)
    tell_user("The state of your work on this puzzle is represented by ")
    cos = CURRENT_PUZZLE.current_op_sequence
    tell_user("  the operator sequence: " + str(cos))
    solution = list(eval(CURRENT_PUZZLE.correct_sequence))
    if cos==solution:
        tell_user("Congratulations! You solved the puzzle.")
        GameTester.handle_event("Solved puzzle", str(PUZZLE_NUMBER))
        tell_user("Here is a new magic phrase for you:")
        tell_user("    "+CURRENT_PUZZLE.message)
        CURRENT_PUZZLE.solved = True
        GameTester.MAGIC_PHRASE_INVENTORY.append(CURRENT_PUZZLE.message)
        GameTester.MAGIC_PHRASE_PUZZLE_NUM.append(PUZZLE_NUMBER)
        INPUT_ACTIVITY = "navigation"
        main_prompt()
        return

def wall_desc_str(wall):
  #cri = GameTester.CURRENT_ROOM_INDEX
  cri = PRIMEClasses.CURRENT_ROOM_INDEX
  cr = PRIMEClasses.PRIME_ROOMS[str(cri)]
  name = cr.name
  #cfd = xlate_dir(GameTester.CURRENT_FACING_DIRECTION)
  cfd = xlate_dir(PRIMEClasses.CURRENT_FACING_DIRECTION)
  return "This is the "+cfd+" wall in Room "+str(cri)+" ("+name+")."

def xlate_dir(dir):
  if dir=="N": return "North"
  if dir=="E": return "East"
  if dir=="W": return "West"
  if dir=="S": return "South"
  return "(Unknown Direction)"

NOTE_NUMBERS = {'A': 6, 'B': 7, 'C': 0, 'D': 1, 'E': 2, 'F': 3, 'G': 4 }
def xlate_note(note):
  try:
    n = NOTE_NUMBERS[note]
  except:
    n = 8
  return n

# Set up the sound-playing functionality:
# (For windows platforms only, at this time):
from winsound import Beep
h=half_step = pow(2, 1.0/12)
w=whole_step = half_step*half_step
middle_C = 261.63 # (cycles per second)
diatonic_steps = [w,w,h,w,w,w,h]
# Recursively compute a list of diatonic scale ratios:
diatonic_ratios = [1.0] # We start with the ratio for middle C.
for s in diatonic_steps: diatonic_ratios.append(s*diatonic_ratios[-1])
#print "diatonic ratios:", diatonic_ratios
C_scale = [int(middle_C * r) for r in diatonic_ratios]

NOTE_DURATION = 500
def play_notes(note_list):
    index_list = [xlate_note(note) for note in note_list]
    for i in index_list:
        if i > 7: continue
        Beep(C_scale[i], NOTE_DURATION)

# Constants for displaying the maze of rooms:
MAZE_LEFT_MARGIN = 15
MAZE_TOP_MARGIN = 40
ROOM_WIDTH = 120
HALF = ROOM_WIDTH/2
DELTA = ROOM_WIDTH/8
# Methods for displaying the maze of rooms:
MAZE_CANVAS_ITEMS = []
def display_maze():
    #print "Displaying maze..."
    # Adds the game floor plan to the maze canvas at the right place.
    global MZ_CANVAS, PRIMEClasses
    # Clear any previous view:
    for i in MAZE_CANVAS_ITEMS:
        MZ_CANVAS.delete(i)
    # Now draw the current view:
    rooms = PRIMEClasses.PRIME_ROOMS
    #print "rooms is " + str(rooms)
    #x = 20; y = 20; rw = 80; half = 40; delta = 10
    for i in range(3):
        for j in range(3):
            try:
              r = rooms[str(i*3+j)]
            except: continue
            #print "The room at ["+str(i)+", "+str(j)+"] is "+str(r)
            display_room_in_maze(MAZE_LEFT_MARGIN+j*ROOM_WIDTH,
                                 MAZE_TOP_MARGIN+i*ROOM_WIDTH, r)
    CR = PRIMEClasses.CURRENT_ROOM_INDEX
    #print "Current room index is: "+str(CR)
    i = CR / 3; j = CR % 3
    #print "[i, j]  = ["+str(i)+", "+str(j)+"]"
    cx = MAZE_LEFT_MARGIN + ((j)+0.5)*ROOM_WIDTH
    cy = MAZE_TOP_MARGIN  + ((i)+0.5)*ROOM_WIDTH
    FD = PRIMEClasses.CURRENT_FACING_DIRECTION
    #print "Current facing direction is "+FD
    fd = ['N', 'E', 'S', 'W'].index(FD)
    if fd==0: arrow = MZ_CANVAS.create_line(cx, cy + DELTA, cx, cy - DELTA, arrow="last")
    if fd==1: arrow = MZ_CANVAS.create_line(cx - DELTA, cy, cx + DELTA, cy, arrow="last")
    if fd==2: arrow = MZ_CANVAS.create_line(cx, cy + DELTA, cx, cy - DELTA, arrow="first")
    if fd==3: arrow = MZ_CANVAS.create_line(cx - DELTA, cy, cx + DELTA, cy, arrow="first")
    MAZE_CANVAS_ITEMS.append(arrow)
    
def display_room_in_maze(x, y, room):
    global MZ_CANVAS, MAZE_CANVAS_ITEMS, HALF
    rw = ROOM_WIDTH # room width
    item = MZ_CANVAS.create_rectangle(x, y, x+rw-1, y+rw-1, fill="red")
    MAZE_CANVAS_ITEMS.append(item)
    x0 = x; y0 = y; x1=x0; x2 = x+rw*0.3; x3=x2; y1 = y+rw;
    y2 = y+rw*0.7 ; y3 = y+rw*0.3
    item = MZ_CANVAS.create_polygon(x0,y0,x1,y1,x2,y2,x3,y3, fill="blue", outline="black")
    MAZE_CANVAS_ITEMS.append(item)

    x0 = x; y0 = y+rw; x1=x+rw; x2 = x+rw*0.7; x3=x+rw*0.3;
    y1 = y+rw; y2 = y+rw*0.7; y3 = y2
    item = MZ_CANVAS.create_polygon(x0,y0,x1,y1,x2,y2,x3,y3, fill="green", outline="black")
    MAZE_CANVAS_ITEMS.append(item)

    x0 = x+rw; y0 = y+rw; x1=x0; y1=y; x2 = x+rw*0.7; y2 = y+rw*0.3;
    x3=x2; y3 = y+rw*0.7;
    item = MZ_CANVAS.create_polygon(x0,y0,x1,y1,x2,y2,x3,y3, fill="yellow", outline="black")
    MAZE_CANVAS_ITEMS.append(item)

    x0 = x; y0 = y; x1=x+rw; y1=y; x2 = x+rw*0.7; y2 = y+rw*0.3;
    x3=x+rw*0.3; y3 = y2;
    item = MZ_CANVAS.create_polygon(x0,y0,x1,y1,x2,y2,x3,y3, fill="orange", outline="black")
    MAZE_CANVAS_ITEMS.append(item)
    # Display the name of the room
    item = MZ_CANVAS.create_text(x+HALF, y+HALF+15, text=room.name, fill="light blue")
    MAZE_CANVAS_ITEMS.append(item)
    # Display the features of the walls in the room
    for direc in ['N','E','S','W']:
        display_wall_desc_in_maze(room, direc, x + rw/2, y + rw/2)

def display_wall_desc_in_maze(room, loc, xr, yr):
    wall = room.walls[loc]
    #print "wall is: "+str(wall)
    #print "loc is "+loc
    d = ['N', 'E', 'S', 'W'].index(loc)
    dx = get_dx(d)
    dy = get_dy(d)
    desc = wall_desc(wall)
    #print "   wall_desc for "+loc+" is "+desc
    item = MZ_CANVAS.create_text(xr+dx, yr+dy, text=desc)
    MAZE_CANVAS_ITEMS.append(item)

def get_dx(d):
  dx = ROOM_WIDTH / 3 
  if d==0: return 0
  if d==2: return 0
  if d==1: return dx
  if d==3: return -dx
  return 400

def get_dy(d):
  dy = ROOM_WIDTH / 3 
  if d==0: return -dy
  if d==2: return dy
  if d==1: return 0
  if d==3: return 0
  return 400

def wall_desc(w):
  # Return a string that describes key features of the wall.
  desc = ""
  if w.door!=None:
    if w.door.open: desc+='D'
    else: desc+='d'
    desc+=":("+str(w.door.num)+")"
  if w.puzzle_num!=None:
    puzzle = PRIMEClasses.get_puzzle(w.puzzle_num)
    n = int(w.puzzle_num)
    if n>100:
      if puzzle.solved: desc+="M:("+str(n-100)+")\n"+abbrev(puzzle.name)
      else: desc+="m:("+str(n-100)+")\n"+abbrev(puzzle.name)
    else:
      if puzzle.solved: desc+="P:("+str(n)+")\n"+abbrev(puzzle.name)
      else: desc+="p:("+str(n)+")\n"+abbrev(puzzle.name)
  if w.wallpaper!=None: desc+="\n"+abbrev(w.wallpaper)
  return desc

def abbrev(string):
  # If string is not too long, return it intact.
  if len(string)<7: return string
  # Otherwise, return the first few characters of it.
  else: return string[:6]
  
# Prepare for the next line of user input:
def graphicGTInput(prompt):
    global INPUT_PANE
    INPUT_PANE.delete(0, END)
    GameTester.announce(prompt)

RULE_PANE_ITEMS = []
SCORE_ITEM = None
SCORE_FONT = None
FLAG_FONT = None
def show_score():
    global SCORE_ITEM, WV_CANVAS, SCORE_FONT
    if not SCORE_FONT:
      import tkFont
      SCORE_FONT = tkFont.Font(family="Helvetica",size=24)
    if SCORE_ITEM != None: WV_CANVAS.delete(SCORE_ITEM)
    SCORE_ITEM = WV_CANVAS.create_text(180, 420, font=SCORE_FONT, text="Current score: "+\
                                  str(GameTester.SCORE), fill="red")
    refresh_flag_display()

def refresh_flag_display():
    global RULE_PANE, RULE_PANE_ITEMS, FLAG_FONT
    print "In GraphicGameTester.refresh_flag_display, FLAGS is "+str(GameTester.FLAGS)
    if not FLAG_FONT:
      import tkFont
      FLAG_FONT = tkFont.Font(family="Helvetica",size=10)
    for i in RULE_PANE_ITEMS:
      RULE_PANE.delete(i)
    flags = GameTester.FLAGS.keys()
    y = 20
    if flags==[]:
      item = RULE_PANE.create_text(20, y, font=FLAG_FONT,
                                    text="There are no flags currently activated by\n the rules for this game.",
                                    fill="black", anchor="nw")
      RULE_PANE_ITEMS.append(item)
      return
    for flag in flags:
      state = GameTester.FLAGS[flag]
      if state: fill_color = "dark green"
      else: fill_color = "red"
      item = RULE_PANE.create_text(80, y, font=FLAG_FONT,
                                    text="Flag: "+flag+" is "+str(state)+".",
                                    fill=fill_color, anchor="nw")
      RULE_PANE_ITEMS.append(item)
      y += 15

# Replace some GameTester.py functions with new ones: 
GameTester.string_input = graphicGTInput
GameTester.maybe_utter = GGT_maybe_utter
GameTester.do_puzzle_activity = GGT_do_puzzle_activity
tell_user = GameTester.tell_user
GameTester.refresh_flag_display = refresh_flag_display
GameTester.change_room = change_room
GameTester.play_audio_file = play_audio_file

def play_game_from_xml(games_folder, filename, tkroot):
  print "In GraphicGameTester.py, play_game_from_xml, "
  print "  games_folder = "+games_folder
  print "  filename = "+filename
  this_folder = filename[:-4].split("/")[-1]
  folder = games_folder+"/"+this_folder
  #PRIMEClasses.FOLDER = games_folder+"/"+this_folder
  GameTester.load_game(folder, filename=filename)
  #photoimage_test(tkroot)
  setup_and_run(tkroot)

def photoimage_test(tkroot):
  """This method creates a PhotoImage instance and puts in on a canvas.
     It didn't work when called from PRIME-Designer, and this
     same phenomenon was blocking the playing of games when launched from
     PRIME-Designer. The trick is to use the TStar.TK_ROOT when calling
     this from PRIME-Designer."""
  
  frame = Toplevel(tkroot) #, title=" PhotoImage test")
  WV_CANVAS = Canvas(frame, bg='blue', width=500, height=400)
  WV_CANVAS.grid(row=0,column=0,columnspan=2,sticky='news')
  WV_CANVAS.create_text(180,10,text="PhotoImage Test")
  fn="Puzzle-1-img-ver--.jpg"
  get_image(fn)
  WV_CANVAS.create_image(WALL_VIEW_WIDTH/2,WALL_VIEW_HEIGHT/2,
                            image=PHOTOIMAGES[fn],anchor=CENTER)
  tkroot.mainloop()
  
  
if __name__ == '__main__':
  play_audio_file("C:/Documents and Settings/Tanimoto/My Documents/PRIME-Stuff/samplePrime/chimes.wav")
  # Run a sample PRIME game:
  #import os
  #os.chdir("GameFiles")
  #PRIMEClasses.load_game("test_2")
  #GameTester.load_game("Babbage-PRIME")
  #GameTester.load_game(None, filename="C:/Documents and Settings/Tanimoto/My Documents/PRIME-Stuff/samplePrime/samplePrime.xml")
  import os
  os.chdir("C:/Documents and Settings/Tanimoto/My Documents/PRIME-Stuff")
  PRIMEClasses.load_game("samplePrime")
  setup_and_run()
  #TK_ROOT = Tk(className="PhotoImage test")
  #photoimage_test(TK_ROOT)
  
