'''ImagePuzzleDesigner05.py
  This is a component state-space for PRIMEDesigner05.py.
  Since most of the functionality of this module is also needed
  by MusicPuzzleDesigner.py, there are classes in that file that
  subclass those here.
  
  Status as of 23 October 2007: Working.
  
  Includes a group of operators for selecting an initial image and
   secret message.  The precondition should be that the currently
   selected puzzle has nothing in it.

  The message is now nicely embedded in the image.
  The users can enter multiline messages by putting "\n" explicitly
  into the string.
  The message is now rendered with white color.  Perhaps we need one
  more option to select from a set of colors (white, black, red, blue, green)

  I have fleshed out the incorporation of images with the following steps:
   Built the method that loads, constructs and caches images on demand from
   their filenames and size info.  The operator sequence is computed when needed.
   Images are cached at two levels: BASEIMAGES and PHOTOIMAGES.
   The former contains images that have been loaded, resized to a standard
   size, drawn into with the message text, and transformed by shuffling etc.
   The latter have been (a) further resized according to the currently needed
   scale for the view, and (b) converted to the PhotoImage class for use in
   Tkinter.

   The original, possibly large, images directly from files are not cached
   in memory, although it could conceivably be faster sometimes if they were.
   However, there is a risk of using up memory, and we reduce this risk.

   Written by S. Tanimoto as part of the NSF-funded project on
   "Problem-solving methodology in collaborative design."

   (C) 2007 by the University of Washington and Steven Tanimoto.
   
'''

import Image as PIL
import TStar09 as TStar
import COStar09 as COStar
from Tkinter import Canvas
from ImageTk import PhotoImage
import TextListListDesigner05 as TextListListDesigner
import ImageDraw, ImageFont
import primeIMAGEops

GROUP_NAME = None
OBJECT_NAME = None

SHOW_IMAGES = True

# The ImagePState encapsulates a set of image puzzles.
# Most of the representation comes from the TextListListDesigner class
# TextListListState.
# Each puzzle is represented by one TextListWithSelection object.
# The first three lines of text have special significance:
# Name of puzzle, source image, and hidden message.
# The other lines are the names of the transformations that have been
# used to transform the source image (with the message drawn on it) to
# get the scrambled image that serves as the initial state of the puzzle.
class ImagePState(TextListListDesigner.TextListListState):
  def __init__(self,data, parent, last_operator, desc=""):
      super(ImagePState,self).__init__(data, parent, last_operator, desc)
      self.obj_name = OBJECT_NAME
      self.starting_file = "default_image.gif"
  def get_background_color(self): return "red"

class ImagePNode(TextListListDesigner.TextListListNode):
  def paint_content_in_rectangle(self, a_canvas, xc, yc, w, h):
      global SHOW_IMAGES
      if SHOW_IMAGES:
        textlist_y_coord = yc-h/4
        textlist_h = h/2
      else:
        textlist_y_coord = y
        textlist_h = h
      super(ImagePNode,self).paint_content_in_rectangle(a_canvas, xc, textlist_y_coord,
                                                        w, textlist_h,
                                                        show_line_numbers=True)
      sel = self.s.data.selected
      if SHOW_IMAGES and sel>-1:
        tl = self.s.data.tll[sel]
        filename=tl[1][12:]
        #print "In paint_content_in_rectangle, filename is "+filename
        op_sequence=self.build_op_sequence(tl)
        msg = tl[2][16:]
        im = self.get_photoimage(filename,op_sequence,msg,
                                 maxwidth=int(w), maxheight=int(h/2))
        self.p.append(a_canvas.create_image(xc, yc+h/4,image=im, ))

  def build_op_sequence(self, textlist):
    # This should create a tuple representing the sequence of operators chosen,
    # but for now, return an arbitrary code.
    n = len(textlist)
    sequence = []
    for i in range(3,n): # The first 3 lines are name, file, msg (not operators)
      op_index = self.do_map(textlist[i])
      sequence.append(op_index)
    #print "op sequence is: "+str(sequence)
    return tuple(sequence)

  def do_map(self, op_name):
    # Look up the integer index for this operator name.
    # The list of names is different in Image puzzles and Music puzzles,
    # hence this is a class-specific method.
    return MAP_OPNAME_TO_INT[op_name]

  def get_photoimage(self, filename, op_sequence, message,
               maxwidth=128, maxheight=128):
    # Either (1) retrieve, or (2) construct and cache, the needed image.
    # We try a hashtable get, and if nothing there, generate the image,
    global PHOTOIMAGES
    if filename=="Default-image.gif": filename = DEFAULT_IMAGE
    #print "Trying file name: "+filename
    key = filename+"::"+str(maxwidth)+":"+str(maxheight)+":"+str(op_sequence)+message
    try: pim = PHOTOIMAGES[key]
    except:
      im = self.get_baseimage(filename, op_sequence, message)
      if im==None: return None
      im.thumbnail((maxwidth,maxheight))
      pim = PhotoImage(im)
      PHOTOIMAGES[key]=pim
    return pim

  def get_baseimage(self, filename, op_sequence, message, inverse=False):
    global BASEIMAGES, BASEIMAGE_SIZE, FONT, DEFAULT_IMAGE
    if filename=="Default-image.gif": filename = DEFAULT_IMAGE
    if filename=="Default-music-image.gif": filename = DEFAULT_MUSIC_IMAGE
    bkey = filename+"::"+str(op_sequence)+message # size-independent key
    if inverse: bkey+="-INVERSE"
    try:
      im = BASEIMAGES[bkey].copy()
    except:
      try:
        im = PIL.open(filename)
      except:
        print "Could not find, or could not create an image from file with name: "+filename
        return None
      extension = filename[-3:]
      if extension=="GIF" or extension=="gif":
        im = im.convert("RGB")
        print "Converting "+filename+" to RGB"
      message = self.permute_message_if_music(message, op_sequence)
      im.thumbnail((BASEIMAGE_SIZE,BASEIMAGE_SIZE))
      actual_size=im.size
      draw = ImageDraw.Draw(im)
      msglines = message.split("\\n")
      #print "msglines = "+str(msglines)
      nlines = len(msglines)
      linespacing = int(MESSAGE_TYPE_SIZE*1.3)
      #centerx = BASEIMAGE_SIZE/2
      centerx = actual_size[0]/2
      centery = actual_size[1]/2
      ystart = centery - int((nlines/2.0)*linespacing)
      for i in range(nlines):
        size = draw.textsize(msglines[i], font=FONT)
        draw.text((centerx-size[0]/2, ystart+i*linespacing), msglines[i], font=FONT)
      # Now apply the sequence of operators to this image.
      if inverse:
        im = inverse_transform_image(im, op_sequence)
      else:
        im = self.transform_image(im, op_sequence)
      BASEIMAGES[bkey]=im.copy() # Save for future use
    return im
  
  def permute_message_if_music(self, msg, op_sequence):
    return msg # Identity for image puzzles

  def transform_image(self, im, op_sequence):
    for op_idx in op_sequence:
      print "Applying operator "+str(op_idx)+"("+TRANSFORMATION_NAMES[op_idx]+")"
      op_function = TRANSFORMATION_FUNCTIONS[op_idx]
      im = op_function(im)
    return im

#  def inverse_transform_image(self, im, op_sequence):
#    rev_op_seq = op_sequence[:]; rev_op_seq.reverse()
#    for op_idx in rev_op_seq:
#      #print "Applying operator "+str(op_idx)+"("+TRANSFORMATION_NAMES[op_idx]+")"
#      op_function = TRANSFORMATION_FUNCTIONS[op_idx]
#      inv_function = primeIMAGEops.INVERSE[op_function]
#      im = inv_function(im)
#    return im

MESSAGE_TYPE_SIZE = 24
FONT = ImageFont.truetype("arial.ttf", MESSAGE_TYPE_SIZE)

BASEIMAGES = {}  # Images at the "base resolution", some transformed, some not. 
PHOTOIMAGES = {} # Images already resized and converted for display on the canvas.
BASEIMAGE_SIZE = 256

#DEFAULT_IMAGE = "c:/Python25/TStar08/mona-rgb.jpg"
#DEFAULT_IMAGE = "mona-rgb.jpg"
DEFAULT_IMAGE = "common-images/mona-rgb.jpg"
DEFAULT_MUSIC_IMAGE = "common-images/Beethoven-by-Stieler1820.jpg"

horiz_flip_string = "Horizontal flip"
shuffle_rows_string = "Shuffle rows twice"
shuffle_columns_string = "Shuffle columns twice"
rotate_180_string = "Rotate 180"
pixel_crossover_string = "Pixel crossover"

# The following names are used not only in operator menu(s),
# but by the XML conversion method, which works by mapping
# Puzzle transformation descriptions (strings) back to integers.

TRANSFORMATION_NAMES=[horiz_flip_string, shuffle_rows_string,
                      shuffle_columns_string, rotate_180_string,
                      pixel_crossover_string]
MAP_OPNAME_TO_INT = {}
for i in range(len(TRANSFORMATION_NAMES)): MAP_OPNAME_TO_INT[TRANSFORMATION_NAMES[i]]=i
TRANSFORMATION_FUNCTIONS=[primeIMAGEops.horiz_flip,
                          primeIMAGEops.shuffle_rows_twice,
                          primeIMAGEops.shuffle_columns_twice,
                          primeIMAGEops.rotate_180,
                          primeIMAGEops.pixel_crossover]

INVERSE_FUNCTIONS=[primeIMAGEops.horiz_flip,
                          primeIMAGEops.shuffle_rows_twice_inverse,
                          primeIMAGEops.shuffle_columns_twice_inverse,
                          primeIMAGEops.rotate_180,
                          primeIMAGEops.pixel_crossover_inverse]

# The following methods do not perform image operations (those functions are
# in primeIMAGEops.py) but merely insert the NAMEs of operations into
# the list representing the current puzzle.
def horiz_flip(iplws, ps=""):
  iplws2 = TextListListDesigner.add_string_to_selected(iplws,horiz_flip_string)
  return iplws2

def shuffle_image_rows(iplws, ps=""):
  iplws2 = TextListListDesigner.add_string_to_selected(iplws,shuffle_rows_string)
  return iplws2

def shuffle_image_columns(iplws, ps=""):
  iplws2 = TextListListDesigner.add_string_to_selected(iplws,shuffle_columns_string)
  return iplws2
  
def rotate_180(iplws, ps=""):
  iplws2 = TextListListDesigner.add_string_to_selected(iplws,rotate_180_string)
  return iplws2

def pixel_crossover(iplws, ps=""):
  iplws2 = TextListListDesigner.add_string_to_selected(iplws,pixel_crossover_string)
  return iplws2
  
#def replace_starting_image(iplws, filename, applic_time_param_str=""):
#  iplws2 = TextListListDesigner.replace_first_line(iplws, "Start with: "+filename)
#  return iplws2

def setup():
  global GROUP_NAME, OBJECT_NAME
  GROUP_NAME = "Image Puzzle Designer"
  OBJECT_NAME = "image puzzle"
  basic_ops = TextListListDesigner.make_basic_tll_ops(obj_name=OBJECT_NAME,
                                                    group_name=GROUP_NAME,
                                                    tll_type="ImageP")
  image_op1 = TStar.OperatorSpecifiedWithDialog("Replace puzzle name",
                           TextListListDesigner.replace_puzzle_name,
                           group=GROUP_NAME,
                           precondition=TextListListDesigner.need_selection,
                           dialog_title="Enter the new name",
                           prompt="Puzzle name",
                            )
  image_op2 = TStar.OperatorSpecifiedWithDialog("Replace starting image",
                           #replace_starting_image,
                           TextListListDesigner.replace_starting_file,
                           group=GROUP_NAME,
                           precondition=TextListListDesigner.need_selection,
                           dialog_title="Choose a file",
                           prompt="Select the image file of your choice",
                           param_op_type="File",
                           filter_desc=[('JPG files','*.JPG'),
                                        ('GIF files','*.GIF'),
                                        ('PNG files','*.PNG'),
                                        ('All files','*')])
  image_op3 = TStar.OperatorSpecifiedWithDialog("Replace embedded message",
                           TextListListDesigner.replace_embedded_message,
                           group=GROUP_NAME,
                           precondition=TextListListDesigner.need_selection,
                           dialog_title="Enter the hidden message",
                           prompt="Hidden message (one line of text)",
                            )
  image_op4 = TStar.Operator("Horizontal flip", horiz_flip,
                           group=GROUP_NAME,
                           precondition=TextListListDesigner.need_selection)
  image_op5 = TStar.Operator("Shuffle rows", shuffle_image_rows,
                           group=GROUP_NAME,
                           precondition=TextListListDesigner.need_selection)
  image_op6 = TStar.Operator("Shuffle columns", shuffle_image_columns,
                           group=GROUP_NAME,
                           precondition=TextListListDesigner.need_selection)
  image_op7 = TStar.Operator("Rotate by 180", rotate_180,
                           group=GROUP_NAME,
                           precondition=TextListListDesigner.need_selection)
  image_op8 = TStar.Operator("Apply Pixel Crossover", pixel_crossover,
                           group=GROUP_NAME,
                           precondition=TextListListDesigner.need_selection)
  basic_ops += [image_op1, image_op2, image_op3, image_op4, image_op5,
                image_op6, image_op7, image_op8]

  COStar.register_state_space_component(GROUP_NAME,ImagePState, ImagePNode,
                                      basic_ops,
                                      TextListListDesigner.initializer)

  global PHOTOIMAGES
  PHOTOIMAGES = {}
  





