'''MusicPuzzleDesigner05.py.
   This file provides the music puzzle components space elements
   for PRIMEDesigner.
   A music puzzle is a special kind of image puzzle that has
   musical elements in addition to the image elements.

  Written by S. Tanimoto as part of the NSF-funded project on
  "Problem-solving methodology in collaborative design."

  (C) 2007 by the University of Washington and Steven Tanimoto.

'''

import TStar09 as TStar
import COStar09 as COStar
import ImageP.ImagePuzzleDesigner05 as ImagePD
import TextListListDesigner05 as TextListListDesigner
from Tkinter import Canvas

GROUP_NAME = None
OBJECT_NAME = None

class MusicPState(ImagePD.ImagePState):
  def __init__(self,data, parent, last_operator, desc=""):
      super(ImagePD.ImagePState,self).__init__(data, parent, last_operator, desc)
      self.obj_name = OBJECT_NAME
  def get_background_color(self): return "white"

class MusicPNode(ImagePD.ImagePNode):
  def do_map(self, op_name):
    global MAP_OPNAME_TO_INT
    return MAP_OPNAME_TO_INT[op_name]

  def permute_message_if_music(self, msg, op_sequence):
    return permute_notes(msg, op_sequence)

  def transform_image(self, im, op_sequence): return im # No image manip for music puzzle.

def permute_notes(msg, op_sequence):
  for op_idx in op_sequence:
    print "Applying operator "+str(op_idx)+"("+TRANSFORMATION_NAMES[op_idx]+")"
    op_function = TRANSFORMATION_FUNCTIONS[op_idx]
    msg = op_function(msg)
  return msg
  
def reverse_string(s):
  s2 = ""
  for c in s:
    s2 = c + s2
  return s2

def shuffle_string(s):
  s2 = ""
  n = len(s)
  if n%2==0: s+=' '; n+=1 # Pad to next odd number of chars for invertibility
  for i in range(n):
    s2 += s[(2*i)%n]
  return s2

def odd(n): return (n % 2)==1

def unshuffle_string(s):
  s2 = ""
  n = len(s)
  #if n%2==0: s+=' '; n+=1 # Pad to next odd number of chars for invertibility
  for i in range(n):
    if odd(i):
      i0 = n-((n-i+1)/2)
    else: i0 = i/2
    s2 += s[i0]
  return s2

def reflect_pitches_in_string(s):
  s2=""
  for c in s:
    try: newc = {'A':'G', 'B':'F', 'C':'E', 'D':'D', 'E':'C', 'F':'B', 'G':'A'}[c]
    except: newc = c
    s2 += newc
  return s2

INVERSE = {reverse_string: reverse_string,
           shuffle_string: unshuffle_string,
           reflect_pitches_in_string: reflect_pitches_in_string}

# The following names are used not only in operator menu(s),
# but by the XML conversion method, which works by mapping
# Puzzle transformation descriptions (strings) back to integers.
reverse_tune_string = "Reverse tune"
shuffle_notes_string = "Shuffle notes"
reflect_pitches_string = "Reflect pitches"
TRANSFORMATION_NAMES=[reverse_tune_string, shuffle_notes_string, reflect_pitches_string]
MAP_OPNAME_TO_INT = {}
for i in range(len(TRANSFORMATION_NAMES)): MAP_OPNAME_TO_INT[TRANSFORMATION_NAMES[i]]=i
TRANSFORMATION_FUNCTIONS=[reverse_string,
                          shuffle_string,
                          reflect_pitches_in_string]
INVERSE_FUNCTIONS=[reverse_string,
                          unshuffle_string,
                          reflect_pitches_in_string]

def reverse_tune(mplws, ps=""):
  mplws2 = TextListListDesigner.add_string_to_selected(mplws,reverse_tune_string)
  return mplws2

def shuffle_notes(mplws, ps=""):
  mplws2 = TextListListDesigner.add_string_to_selected(mplws,shuffle_notes_string)
  return mplws2

def reflect_pitches(mplws, ps=""):
  mplws2 = TextListListDesigner.add_string_to_selected(mplws,reflect_pitches_string)
  return mplws2

def setup():
  global GROUP_NAME, OBJECT_NAME
  GROUP_NAME = "Music Puzzle Designer"
  OBJECT_NAME = "music puzzle"
  basic_ops = TextListListDesigner.make_basic_tll_ops(obj_name=OBJECT_NAME,
                                                    group_name=GROUP_NAME,
                                                    tll_type="MusicP")
  music_op1 = TStar.Operator("reverse tune", reverse_tune,
                           group=GROUP_NAME,
                           precondition=TextListListDesigner.need_selection)
  music_op2 = TStar.Operator("shuffle notes", shuffle_notes,
                           group=GROUP_NAME,
                           precondition=TextListListDesigner.need_selection)
  music_op3 = TStar.Operator("reflect pitches", reflect_pitches,
                           group=GROUP_NAME,
                           precondition=TextListListDesigner.need_selection)
  music_op5 = TStar.OperatorSpecifiedWithDialog("Replace puzzle name",
                           TextListListDesigner.replace_puzzle_name,
                           group=GROUP_NAME,
                           precondition=TextListListDesigner.need_selection,
                           dialog_title="Enter the new name",
                           prompt="Puzzle name")
  
  music_op4 = TStar.OperatorSpecifiedWithDialog("Replace background image",
                           TextListListDesigner.replace_starting_file,
                           group=GROUP_NAME,
                           precondition=TextListListDesigner.need_selection,
                           dialog_title="Choose a file",
                           prompt="Select the image file of your choice",
                           param_op_type="File",
                           filter_desc=[('JPG files','*.JPG'),
                                        ('GIF files','*.GIF'),
                                        ('PNG files','*.PNG'),
                                        ('BMP files','*.BMP'),
                                        ('All files','*')])
  
  music_op6 = TStar.OperatorSpecifiedWithDialog("Replace hidden melody/message",
                           TextListListDesigner.replace_embedded_message,
                           group=GROUP_NAME,
                           precondition=TextListListDesigner.need_selection,
                           dialog_title="Enter the message (Use only letters ABCDEFG",
                           prompt="New melody-message")

  basic_ops += [music_op5, music_op4, music_op6, music_op1, music_op2, music_op3]

  COStar.register_state_space_component(GROUP_NAME,MusicPState, MusicPNode,
                                      basic_ops,
                                      TextListListDesigner.initializer)
    

# Stuff for playing music:
NOTE_NUMBERS = {'A': 5, 'B': 6, 'C': 0, 'D': 1, 'E': 2, 'F': 3, 'G': 4 }
def xlate_note(note):
  try:
    n = NOTE_NUMBERS[note]
  except:
    n = 8
  return n

# Set up the sound-playing functionality:
# (For windows platforms only, at this time):
from winsound import Beep
h=half_step = pow(2, 1.0/12)
w=whole_step = half_step*half_step
middle_C = 261.63 # (cycles per second)
diatonic_steps = [w,w,h,w,w,w,h]
# Recursively compute a list of diatonic scale ratios:
diatonic_ratios = [1.0] # We start with the ratio for middle C.
for s in diatonic_steps: diatonic_ratios.append(s*diatonic_ratios[-1])
#print "diatonic ratios:", diatonic_ratios
C_scale = [int(middle_C * r) for r in diatonic_ratios]

NOTE_DURATION = 500
def play_notes(note_list):
    index_list = [xlate_note(note) for note in note_list]
    for i in index_list:
        if i > 7: continue
        Beep(C_scale[i], NOTE_DURATION)

def play_current_melody(music_puzzles):
  if len(music_puzzles.tll)<1:
      import tkMessageBox
      tkMessageBox.showwarning("The selected state has no music puzzles.\n")
      return
  if music_puzzles.selected < 0:
      tkMessageBox.showwarning("No music puzzle is selected in the selected state.\n")
      import tkMessageBox
      return    
  mp = music_puzzles.tll[music_puzzles.selected]
  notes = mp[2]
  dummyNode = MusicPNode(music_puzzles)
  op_sequence = dummyNode.build_op_sequence(mp)
  permuted_notes = permute_notes(notes, op_sequence)
  play_notes(permuted_notes)

DEFAULT_MUSIC_IMAGE = "Image-resources/Beethoven-by-Stieler1820.jpg"





