'''PRIMEDesigner05.py
   This is Version 0.5 of the PRIME Designer program.
   Last updated March 31, 2009, with support for a rudimentary
     "detailed view" of states that can be useful to see the full
     filenames of elements of a game. The view doesn't fully "decode"
     some of the internal representations (like wall codes and
     untranslated puzzle numbers) so could be improved in the future.
   New in 0.5: (a) logging to the local hard drive.
   Fixes made as of 24 March 2009: 
     1. (DONE) can play last saved game, even if saved in an earlier session.
          (It uses a special file called "Last-game-exported.txt" to identify
          which game should be played.)
        (NOT YET) Or can play a named game other than the one being edited,
          via a file dialog.
     2. (DONE) Design session fully recovers from a game engine crash.
     3. (DONE) Menu option to export game without creating puzzle images.
        Very useful when puzzle images were already created in an earlier 
        export operation.
     4. (DONE) more obvious highlighting of the focus node.
        Done by adjusting some code in TStar09.py for node painting of focus node.
        Lowered the threshold from 0.8 to 0.4 for the larger margins, and
        increased larger margin from 6 to 10
     5. (DONE) added a progress meter to the exporting of precomputed
        game images, which is the most time-consuming task we have.

     NOT IMPLEMENTED AT THIS TIME:
        compile-time option to use RELATIVE PATHS for resources
        in the PDT files,
        rather than absolute.

   PURPOSE of PRIME Designer:
   
   PRIME Designer adapts COStar for the integration of the separate
   design spaces for the four media.
   There is a menu command to export a PRIME game as an XML file.

   FEATURES OF V03:
   As of 13 Sept 2007, all BASIC functionality is working:
   Although no music files are exported
     the XML file includes permuted note sequences.
   Features for merging trees have been included.
   Communication works with the server ole.cs.washington.edu.

   17 Sept: I have added most of the code to GraphicGameTester
   and GameTester to automatically play "background audio" associated
   with each room as the player enters the room.  Before this can
   be tested, the following are needed:
   A. (NOW DONE) Code in PRIMEClasses.py to read this attribute from XML and
   store it as the .background_audio instance variable of the ROOM.
   B. DONE. Add a new operator to the Architect's palette to replace the
   audio of the current room.
   C. DONE. Add code to Write this file info in the XML file.
   
   D. DONE. Add a new ACTION to the RULE Designer's palette which
   is to PLAY an audio file. (Nice thing to do after solving a
   music puzzle -- Play a recording of the full, real piece of
   which the puzzle solution is a fragment.)
   E. FIXED by editing the SamplePrime.xml file.
      THERE was a problem showing up with the SamplePrime.xml
   game in My Documents/PRIME-Stuff.  When moving into the INNER_SANCTUM
   the GraphicGameTester tried to load the music files .not file as
   an image file and threw an error.

   FEATURES OF V04:
   7 Feb 2008.  Modifying the way images are represented in pdt files:
   Instead of absolute paths, they will be relative paths.  We will assume
   that there are four folders of resources:
     common-images/
     new-images/
     common-audio/
     new-audio/
   Those files in common-images and in common-audio are assumed to be
   pre-installed on all systems involved in the collaborative design session.
   Those in new-images and new-audio may have to be shared through an
   extra process of communication. For examples, when posting a design file
   that includes references to new files, these may need to be uploaded to
   a session-oriented file area for sharing.
   To implement these changes, TStar08.py has been modified.

   in primeXMLwriter.py, the code does not change the current directory
   any more.  This makes it easier to export games.
   

   Written by S. Tanimoto as part of the NSF-funded project on
   "Problem-solving methodology in collaborative design."

   (C) 2007 by the University of Washington and Steven Tanimoto.
   
'''
import COStar09 as COStar

ADVANCED = True  # Enable means to create customized displays
#ADVANCED = False  # Avoid unnecessary T-Star menus that might confuse
# Users during the PRIME Designer experiments.


# Import and set up each of the 4 component
# design spaces.
# If this order is ever changed, then it will be
# necessary to change the indexing in the method
# make_xml, which is below.
import Arch.ArchitecturalDesigner05 as ArchD
ARCHITECT_ROLE = 0
import ImageP.ImagePuzzleDesigner05 as ImagePD
IMAGE_PUZZLER_ROLE = 1
import MusicP.MusicPuzzleDesigner05 as MusicPD
MUSIC_PUZZLER_ROLE = 2
import RuleB.RuleBaseDesigner05     as RuleBD
RULE_BASE_DESIGNER_ROLE = 3
ALL_ROLES = 4

def setup():
  print "OK"
  COStar.initialize()
  ArchD.setup()
  ImagePD.setup()
  MusicPD.setup()
  RuleBD.setup()

# Create menu item for XML
#from Tkinter import Menu, Toplevel
from Tkinter import *
import tkMessageBox

def make_more_app_menus(menubar):
  xmlMenu = Menu(menubar)
  menubar.add_cascade(label="PRIME Options", menu=xmlMenu)
  xmlMenu.add_command(label="Export draft game folder (low-res puzzle images)", command=make_draft_folder)
  xmlMenu.add_command(label="Export full game folder (full-res puzzle images)", command=make_full_folder)
  xmlMenu.add_command(label="Export but skip puzzle images", command=make_no_images)
  xmlMenu.add_command(label="Write only the XML File", command=make_xml)
  xmlMenu.add_command(label="Play exported game in 2D Graphical env't", command=playgame)
  xmlMenu.add_command(label="Play exported game in 3D Graphical env't", command=playgame3D)
  xmlMenu.add_command(label="Set game properties", command=edit_properties)
  xmlMenu.add_command(label="Play melody of selected node", command=play_melody)
  xmlMenu.add_command(label="Melody suggestions", command=melody_list)
  xmlMenu.add_command(label="Post current design on server", command=post)
  xmlMenu.add_command(label="Get design from server and merge with existing", command=getmerge)
  #xmlMenu.add_command(label="Test progress meter", command=test_meter)
  

import primeXMLwriter05 as primeXMLwriter
import TStar09 as TStar
USE_FOLDER=False
def make_xml(fullfolder=False, draft=False, images=True):
  '''This method prompts the user for a filename and writes
  an XML-format file that describes the PRIME game represented
  by the currently selected T-Star node.
  Note: This is NOT usually the most recently created node,
  but its parent.  If you want the very last-created node
  converted to XML, you will need to explicitly select it with
  the mouse or create a child of it.
  '''
  log("make_xml")
  sn = TStar.TREE_DISP.focus_node
  if not require_focused(sn): return
  current_state = sn.s
  fp = current_state.data[0]
  ip = current_state.data[1].tll
  mp = current_state.data[2].tll
  rb = current_state.data[3].tll
  props = [GAME_TITLE, GAME_VERSION, GAME_AUTHORS]
  global USE_FOLDER
  primeXMLwriter.write_prime_game_as_xml(fp,ip,mp,rb,props,
                                         use_folder=USE_FOLDER,full=fullfolder,
                                         draft=draft,images=images)

def require_focused(sn):
  if sn==None:
      tkMessageBox.showwarning("No node selected", "You must first select a node for XML conversion.\n")
      return False
  return True

def make_full_folder(draft=False, images=True):
  log("make_full_folder")
  if not require_focused: return
  global USE_FOLDER
  USE_FOLDER=True
  make_xml(fullfolder=True, draft=draft, images=images)
  USE_FOLDER=False
  
def make_draft_folder():
  log("make_draft_folder")
  make_full_folder(draft=True)

def make_no_images():
  log("make_no_images")
  make_full_folder(images=False)

def playgame():
  log("playgame")
  game_to_play = determine_game_filename()
  if game_to_play==None: return
  from GraphicGameTester import play_game_from_xml
  play_game_from_xml(primeXMLwriter.GAMES_FOLDER, game_to_play,
                     TStar.TK_ROOT)

def playgame3D():
  log("playgame3D")
  game_to_play = determine_game_filename()
  if game_to_play==None: return
  lst = game_to_play.split("/")
  stripped_name = lst[-2]
  import os
  savedir = os.getcwd()
  os.chdir("GameFiles")
  try: 
    os.system("sod "+stripped_name)
  finally:
    os.chdir(savedir)

def determine_game_filename():
  game_to_play = primeXMLwriter.XML_FILENAME
  if game_to_play==None:
    # try to find out what game was exported most recently in any session.
    try:
      f = open("last_game_exported.txt", "r")
      game_to_play = f.read()
    except:
      tkMessageBox.showwarning("No XML file was created", "You must first create an XML file.\n")
      return None
  print "The game to play is "+game_to_play
  return game_to_play
      
def play_melody():
  log("play_melody")
  sn = TStar.TREE_DISP.focus_node
  if not require_focused(sn): return
  current_state = sn.s
  music_puzzles = current_state.data[2] # Dependent on order of components in PRIME-Designer.
  MusicPD.play_current_melody(music_puzzles)

def post():
  # Upload the design to the server.
  log("post")
  from HTTPPRIMEClient import post_multipart
  # Create a tree description file, an annotation file, and
  # get comments from the user.
  TStar.tree_file_pre_operation() # No-op except in COStar applications (sets ROLE to ALL_ROLES)
  TStar.PROPERTIES["USE_ABSOLUTE_OPERATOR_NOS"] = True
  TStar.USE_ABSOLUTE_OPERATOR_NOS = True
  tree_desc_str = TStar.INITIAL_STATE.get_tree_desc_str()
  TStar.tree_file_post_operation() # (Restores ROLE)
  pdt_header = TStar.HEADLINE+"\n"
  pdt_header+= "###USER="+TStar.USER+"\n"
  import time
  pdt_header+= "###TIME="+str(time.asctime())+"\n"
  pdt_header+= "###VIEW="+TStar.VIEW+"\n"
  pdt_header+= "###PROPERTIES="+str(TStar.PROPERTIES)+"\n"
  pdt_data = pdt_header+tree_desc_str
  # Get the annotations...
  annotations = ""
  # Walk the tree, and at each node, execute a save_node_annotations function
  for s in TStar.ALL_STATES:
      thisNode = TStar.TREE_DISP.state_node_hash[s]
      annotations += thisNode.annotations_to_loadable_str()
  # Get comments:
  from tkSimpleDialog import askstring
  comment = askstring("Enter a comment", "Comment for this version...")

  response = \
  post_multipart("ole.cs.washington.edu", "/htbin-post/PRIME/upload.py",
                 [("username", USERNAME),
                  ("session_name", SESSION_NAME),
                  ("comments", comment)],
                 [("pdtfile", "prime.pdt", pdt_data),
                  ("annfile", "prime.ann", annotations)])
  print "Response is: "+response

SERVER_CHOICES = []

def getmerge():
  log("getmerge")
  global SESSION_NAME
  # Get a list of choices from the server
  print "Preparing to request a list of files from the server."
  try: 
    from HTTPPRIMEClient import post_multipart
    response = post_multipart("ole.cs.washington.edu", "/htbin-post/PRIME/getlist.py",
                   [("session_name", SESSION_NAME),
                    ("comments", "yes")],[])
                    
    print response
    global SERVER_CHOICES
    SERVER_CHOICES = eval(response)
  except:
    import tkFileDialog
    import tkMessageBox
    result = tkMessageBox.showerror("Error", "Could not get the file list from the server.")
    return
  try:
    sfd = ServerFileDialog(TStar.TK_ROOT)
    file_choice = sfd.result
    print file_choice
    i = file_choice.index(':')
    basename = file_choice[:i]
    tree_file_text = get_file_text(basename+".pdt")
    print "THE TREE:\n"+tree_file_text
    annotation_file_text = get_file_text(basename+".ann")
    print "THE ANNOTATIONS:\n"+annotation_file_text
  except:
    import tkFileDialog
    import tkMessageBox
    result = tkMessageBox.showerror("Error", "Could not load the files from the server.")
    return
  TStar.open_tree_file(remove_existing=False, desc_string=tree_file_text)
  COStar.load_annotations(None, desc_string=annotation_file_text)
    

def get_file_text(filename):
  global SESSION_NAME
  from HTTPPRIMEClient import post_multipart
  response = post_multipart("ole.cs.washington.edu", "/htbin-post/PRIME/get-raw-file.py",
                 [("session_name", SESSION_NAME),
                  ("filename", filename)],[])
  return response
  


import tkSimpleDialog
import Tkinter
class ServerFileDialog(tkSimpleDialog.Dialog):

    def body(self, master):
        global SERVER_CHOICES, SESSION_NAME
        self.lb =Tkinter.Listbox(master,selectmode="SINGLE", width=150)
        for filen in SERVER_CHOICES:
          self.lb.insert(Tkinter.END, filen)
        self.lb.pack()
        return self.lb # initial focus

    def apply(self):
        self.selected = self.lb.curselection()
        print self.selected
        self.result = SERVER_CHOICES[int(self.selected[0])]
  

import tkSimpleDialog
from Tkinter import Label, Entry, Listbox, Checkbutton, Canvas, END,\
  IntVar, StringVar, Text
          
def melody_list():
  import tkMessageBox
  tkMessageBox.showinfo("Suggested melody-messages",\
        "ABBACADABBA\n"
        "ABE BADE GABE A BABE\n"
        "A C. BABBAGE ADAGE: ADA ADD DADA\n"
        "ACE ACE ACE ACE\n"
        "ADD A BEAD\n"
        "ADDED BAGGAGE\n"
        "AGED ADAGE\n"
        "BAD DEED\n"
        "BAGGED FEED\n"
        "BEEBEE BE BAD\n"
        "CAB GAB\n"
        "CABBAGE BED\n"
        "CAGED EGG\n"
        "CEDED A DEED\n"
        "DADA CAFE\n"
        "DAD FED AGED BEEF\n"
        "DEAD DAD\n"
        "EEG BAD\n"
        "EFFACED\n"
        "EGAD A BAD FAD\n"
        "FAB DECADE\n"                
        "FADED FACADE\n"                
        "GABBED BAD GAB\n"                
        "GAFFE GAGGED")
  
GAME_TITLE = "Untitled game"
GAME_VERSION = "0.1"
GAME_AUTHORS = "Enter names here"

def edit_properties():
  log("edit_properties")
  prop = PropDialog(TStar.TK_ROOT)
  
class PropDialog(tkSimpleDialog.Dialog):
        
    def body(self, master):
        global GAME_TITLE, GAME_VERSION, GAME_AUTHORS
        self.title("Edit PRIME game properties")
        self.nm = StringVar()
        Label(master, text="Name of game").grid(row=1,column=0)
        nameentry = Entry(master, width=40, textvariable=self.nm)
        nameentry.grid(row=1, column=1)
        self.nm.set(GAME_TITLE)
        self.vs = StringVar()
        Label(master, text="Version number").grid(row=2,column=0)
        versentry = Entry(master, width=40, textvariable=self.vs)
        versentry.grid(row=2, column=1)
        self.vs.set(GAME_VERSION)
        self.au = StringVar()
        Label(master, text="Names of authors").grid(row=3,column=0)
        authentry = Entry(master, width=40, textvariable=self.au)
        authentry.grid(row=3, column=1)
        self.au.set(GAME_AUTHORS)
        Label(master, text="PROPERTIES").grid(row=0,column=0)
    def apply(self):
        global GAME_TITLE, GAME_VERSION, GAME_AUTHORS
        GAME_TITLE = self.nm.get()
        GAME_VERSION = self.vs.get()
        GAME_AUTHORS = self.au.get()
        record_properties()                
        TStar.TK_ROOT.wm_title(TOOL_NAME+': '+GAME_TITLE)

def tree_file_pre_operation():
  global SAVED_ROLE, ALL_ROLES, GAME_TITLE, GAME_VERSION, GAME_AUTHORS
  print "In PRIMEDesigner05.py, saving COStar.ROLE which is "+str(COStar.ROLE)
  print "In PRIMEDesigner05.py, saving GAME_TITLE which is "+GAME_TITLE
  COStar.ROLE=ALL_ROLES
  TStar.PROPERTIES['GAME_TITLE'] = GAME_TITLE
  TStar.PROPERTIES['GAME_VERSION'] = GAME_VERSION
  TStar.PROPERTIES['GAME_AUTHORS'] = GAME_AUTHORS
  
def tree_file_post_operation():
  global SAVED_ROLE, GAME_TITLE, GAME_VERSION, GAME_AUTHORS
  COStar.ROLE = TStar.PROPERTIES['ROLE']
  print "In PRIMEDesigner05.py, restoring COStar.ROLE which is "+str(COStar.ROLE)
  GAME_TITLE=TStar.PROPERTIES['GAME_TITLE']
  GAME_VERSION=TStar.PROPERTIES['GAME_VERSION']
  GAME_AUTHORS=TStar.PROPERTIES['GAME_AUTHORS'] 

def record_properties():
  TStar.PROPERTIES['GAME_TITLE']=GAME_TITLE 
  TStar.PROPERTIES['GAME_VERSION']=GAME_VERSION 
  TStar.PROPERTIES['GAME_AUTHORS']=GAME_AUTHORS


LOGGING_ENABLED = True
LOGGING_FILENAME = None
import time
def date_time_for_filename():
  t = time.localtime()
  l = list(t)
  s = map(str, l)
  r = reduce(lambda x,y:x+'-'+y, s)
  return r
  #t = time.asctime()
  #return t[-13:-11]+'-'+t[-10:-8]+'-'+t[-7:-5] # e.g., 13-26-59 (HH-MM-SS)

def time_for_timestamp():
  t = time.asctime()
  return t[-13:-5]

def log(msg):
  global LOGGING_FILENAME, LOGGING_ENABLED
  if not LOGGING_ENABLED: return
  logging_directory = "logfiles"
  # See if the log file already has been started.
  if LOGGING_FILENAME==None:
    # No, make sure the directory exists.
    import os
    if not os.path.exists(logging_directory):
        os.makedirs(logging_directory)
        print "Created logging directory"

    # And come up with a new filename.
    logID = date_time_for_filename()
    now = time_for_timestamp()
    filename = logging_directory+"/PDLF-"+logID+".txt"
    # Create a new file
    print "Trying to create log file: "+filename
    fp = open(filename, 'w')
    fp.write('# PRIME Designer Log File for session started at' + now + "\n")
    fp.close()
    # Remember this file name.
    LOGGING_FILENAME = filename
  print "Logging: "+msg
  fp = open(LOGGING_FILENAME, 'a')
  fp.write(msg + ' [at '+time_for_timestamp()+']\n')
  fp.close()

def open_detailed_view():
  #set_last_command(show_details)
  this_state = TStar.TREE_DISP.focus_node.s
  top = Toplevel(TStar.TK_ROOT)
  top.grid_rowconfigure(0, weight=1)
  top.grid_columnconfigure(0, weight=1)
  detail_canvas = Canvas(top, width=800, height=1000)
  detail_canvas.grid(row=0,column=0,sticky='nesw')
  state_desc_lst = [substate.describe()+"\n" for substate in this_state.data]
  desc = "Detailed view. Maybe useful for checking filenames.\n\n"
  if COStar.ROLE == 4:
    for ds in state_desc_lst:
      desc += ds
  else:
    desc = state_desc_lst[COStar.ROLE]
    print desc
  #from Tkinter import Canvas
  detail_canvas.create_text(100,30,text=desc,anchor=NW)

# Additional changes to some titles and the tree type.
TOOL_NAME = 'PRIME-Designer, Version 0-5'
TStar.__dict__.update({
    'TITLE':' '+TOOL_NAME,
    'HEADLINE':"###PRIME Designer Version 0.5 Tree Description File",
    'ADVANCED':ADVANCED,
    'TREE_FILE_TYPE':"pdt",
    'log':log,
    'PREFER_ABSOLUTE_OPERATORS':True,
    'open_detailed_view':open_detailed_view,
    'tree_file_pre_operation':tree_file_pre_operation,
    'tree_file_post_operation':tree_file_post_operation})

def set_role(role):
  COStar.ROLE=role
  TStar.PROPERTIES['ROLE']=role

USERNAME = "Joe User"
SESSION_NAME = "default_session"
INITIAL_ROLE = 2

def login():
  from Tkinter import Tk, Label, StringVar, Entry
  temp_TK_ROOT = Tk(className="Log in")
  Label(temp_TK_ROOT, text="Welcome to PRIME Designer").pack()
  prop = LoginDialog(temp_TK_ROOT)
  temp_TK_ROOT.destroy()
  global USERNAME, SESSION_NAME
  #print "USERNAME is now "+USERNAME
  #print "SESSION_NAME is now "+SESSION_NAME
  log("USERNAME is "+USERNAME)
  log("SESSION_NAME is "+SESSION_NAME)
  
  
class LoginDialog(tkSimpleDialog.Dialog):
        
    def body(self, master):
        global USERNAME, SESSION_NAME, INITIAL_ROLE
        self.title("Log in to PRIME Designer")
        self.um = StringVar()
        Label(master, text="User name").grid(row=1,column=0)
        usernameentry = Entry(master, width=40, textvariable=self.um)
        usernameentry.grid(row=1, column=1)
        self.um.set(USERNAME)
        self.sn = StringVar()
        Label(master, text="Session name").grid(row=2,column=0)
        versentry = Entry(master, width=40, textvariable=self.sn)
        versentry.grid(row=2, column=1)
        self.sn.set(SESSION_NAME)
        self.ir = StringVar()
        Label(master, text="Initial role").grid(row=3,column=0)
        authentry = Entry(master, width=40, textvariable=self.ir)
        authentry.grid(row=3, column=1)
        self.ir.set(INITIAL_ROLE)
        Label(master, text="PROPERTIES").grid(row=0,column=0)
    def apply(self):
        global USERNAME, SESSION_NAME, INITIAL_ROLE
        USERNAME = self.um.get()
        SESSION_NAME = self.sn.get()
        INITIAL_ROLE = int(self.ir.get())
        TStar.USER = USERNAME
        log("Initial role: "+str(INITIAL_ROLE))

def test():
  global INITIAL_ROLE
  setup()
  record_properties()
  COStar.make_more_app_menus = make_more_app_menus # Use this version
  login()
  TStar.ZOOM_FACTOR = 0.5
  # print "INITIAL_ROLE is "+str(INITIAL_ROLE)
  set_role(INITIAL_ROLE)
  COStar.go()

test()

