'''TextListListDesigner.py
  This is a generic component state-space for implementing
  three of the specific component state spaces for PRIMEDesigner03.py.
  Each state is a list of lists of strings.
  The ops provide for:
    Adding a new list of strings.
    Selecting an existing list of strings
    Removing the selected list of strings
    Adding a new string to the currently selected list of strings

  Status as of 25 August 2007: Working
     The set of basic operators provided here can be extended by
        child classes for image puzzle design, rule-base design, etc.
     A method here displays the selection CONTEXT (which element is
        selected, and out of how many elements.
        This could be improved to show a box (and number on top of it)
        for each text list, with the selected one in the middle, and the
        total number of text lists shown.
        
     DONE:
        Factor out the display method, so that there can be a display
           of a graphic rep. later -- image or music, etc.
        
     written by S. Tanimoto.
  Last updated 12 April 2009.
  '''

import TStar09 as TStar
import COStar09 as COStar
from Tkinter import Canvas

GROUP_NAME = None
BACKGROUND_COLOR = "gray"

''' Here is a class that encapsulates the data object used in the
states of this (sub)search space.'''
class TextListListWithSelection:
  def __init__(self):
      self.selected = -1 # nothing selected
      self.tll = []      # the actual text list list.
      self.lineToHighlight = -1 # Important for rule editing
  def describe(self):
      desc = ""
      i = 0
      for tl in self.tll:
        desc += "TextList "+str(i)+":\n"
        for t in tl:
          desc += t + "\n"
        i += 1
      return desc

''' Here is a class that represents the (component) state in a
PRIMEDesigner state space.'''
class TextListListState(TStar.State):
  def __init__(self,data, parent, last_operator, desc=""):
      # The following call to 'super' works because this class is
      # defined as a "new Python class".  This is accomplished by
      # a tiny modification to TStar.State in TStar (TStar09.py) that has
      # it subclass 'object' instead of subclassing nothing.
      # Without this change, the program either won't compile or
      # after changing 'super(TextListListState,self)' to
      # 'super(type(TextListListState),self)', compiles but gives
      # a runtime error. (I learned all this the hard way!)
      super(TextListListState,self).__init__(data, parent, last_operator, desc)
      # And here is the point of overriding the parent __init__,
      # to add a field to the state called obj_name.
      self.obj_name = "text list"

  def get_background_color(self): return "light gray"
  
  def paint_content_in_rectangle(self, a_canvas, xc, yc, w, h,
                                 show_line_numbers=False):
      # Does the real work of displaying the text list list.
      # It returns a list of the new canvas items so that the
      # calling method in the node class can append them to the
      # list of items that should be deleted before the node
      # is later redrawn.
      new_items = []
      bg_color = self.get_background_color()
      r = a_canvas.create_rectangle(xc-w/2,yc-h/2,xc+w/2,yc+h/2,fill=bg_color)
      new_items.append(r)
      selection_display_height = 20
      # This code currently only computes rough estimates of the numbers of
      # characters that will fit in the rectangle.
      # At some point, these computations should be recoded to produce more
      # accurate results.
      if h-selection_display_height > 13: # If avail height less than 13, give up on text.
        lineheight = 12 # Later on, compute this from space avail. & prefs. &
        # size of the chosen font or font TO BE chosen.
        nlines = int((h - selection_display_height) / lineheight)
        sel = self.data.selected
        tll = self.data.tll
        lines_in_selected = 0
        line_to_start_with = 0
        if sel > -1:
          lines_in_selected = len(tll[sel])
          if lines_in_selected <= nlines:
            text_to_display = tll[sel] # Enough vert. space for all line.
          else:
            if nlines >= 1:
              line_to_start_with = lines_in_selected - nlines
              text_to_display = tll[sel][line_to_start_with:lines_in_selected] # else show last nlines.
            else:
              text_to_display = ""
          charwidth = 5 # Should be recoded and based on actual font into.
          linelength = int(w / charwidth)
          for row in range(len(text_to_display)):
            if row==self.data.lineToHighlight:
              highlight_height = yc-h/2+selection_display_height+10+row*lineheight-lineheight/2
              r = a_canvas.create_rectangle(xc-w/2,highlight_height,xc+w/2,highlight_height+lineheight,fill="light blue")
              new_items.append(r)
            if show_line_numbers:
              line_number = str(line_to_start_with + row)+': '
            else: line_number = ''
            line_to_display = line_number + text_to_display[row][:linelength]
            t = a_canvas.create_text(xc,yc-h/2+selection_display_height+10+row*lineheight,
                                     text=line_to_display,fill='black')
            new_items.append(t)
      more_items = self.paint_sel_disp_in_rect(a_canvas, xc,
                       yc-h/2+selection_display_height/2,
                       w, selection_display_height)
      new_items += more_items
      return new_items

  def paint_sel_disp_in_rect(self, a_canvas, xc, yc, w, h):
      # Show which text list is selected and out of how many.
      new_items = []
      r = a_canvas.create_rectangle(xc-w/2,yc-h/2,xc+w/2,yc+h/2,fill="pink")
      new_items.append(r)
      item_width = 12
      max_items = int(w / item_width)
      sel = self.data.selected
      n = len(self.data.tll)
      if w < 150: small=True
      else: small=False
      if sel==-1:
        if small: line = "None/"+str(n)
        else: line = str(n)+" "+self.obj_name+"s available. None selected."
      else:
        if small: line = str(sel+1)+"/"+str(n)
        else: line = str(n)+" "+self.obj_name+"s available. "+\
               "#"+str(sel+1)+" selected."
      t = a_canvas.create_text(xc,yc,text=line,fill='black')
      new_items.append(t)
      return new_items
  def describe(self):
      return "A TextListListState object."
    
class TextListListNode(TStar.Node):
  def paint_content_in_rectangle(self, a_canvas, xc, yc, w, h,
                                 show_line_numbers=False):
      new_items = self.s.paint_content_in_rectangle(a_canvas, xc, yc, w, h,
                                                    show_line_numbers)
      self.p += new_items
      
def new_tl(tllws, ps=""):
  "Add a new text list to the list of lists, getting a new TLL."
  tllws2 = TextListListWithSelection() # quasi-clone
  tllws2.tll = tllws.tll[:] # Deep copy not necessary here.
  #if tllws.selected==-1:
  # MAKE THE NEW TL ALWAYS BE THE SELECTED ONE. ADDED 12 APRIL 2009 TO AVOID MERGING PROBLEMS.
  # DURING COLLABORATIVE DESIGN SESSIONS.
  tllws2.selected = len(tllws.tll)
  #else:
  #  tllws2.selected = tllws.selected
  if len(ps)>0:
    tllws2.tll.append(ps)
  else:
    tllws2.tll.append([])
  tllws2.lineToHighlight = -1 
  return tllws2

def new_rule(tllws, ps=""):
  return new_tl(tllws, ["IF:","THEN:","END."])
                
def new_image_puzzle(tllws, ps=""):
  return new_tl(tllws, ["Puzzle name: Untitled image puzzle",
                        "Start with: Default-image.gif",
                        "Hidden message: Open sesame!"])
                
def new_music_puzzle(tllws, ps=""):
  return new_tl(tllws, ["Puzzle name: Untitled music puzzle",
                        "Background: Default-music-image.gif",
                        "Hidden message: EGG FACE"])

DEBUGGING = True                
def add_string_to_selected(tllws, ps="A string"):
  "Add one string to the currently selected list."
  tllws2 = TextListListWithSelection()
  tllws2.tll = tllws.tll[:] # Deep copy not necessary here.
  sel = tllws.selected
  if DEBUGGING:
    print "In TextListListDesigner.add_string_to_selected, sel="+str(sel)
    print "  and the number of TLLs is "+str(len(tllws2.tll))
  newtl = tllws2.tll[sel][:]
  newtl.append(ps)
  tllws2.tll[sel] = newtl
  tllws2.selected = sel
  return tllws2

'''
def replace_first_line(tllws, stored_param, ps=""):
  tllws2 = TextListListWithSelection()
  tllws2.tll = tllws.tll[:] # Deep copy not necessary here.
  sel = tllws.selected
  newtl = tllws2.tll[sel][:]
  newtl[0]=stored_param
  tllws2.tll[sel] = newtl
  tllws2.selected = sel
  return tllws2
'''  

def replace_puzzle_name(tllws, stored_param, ps=""):
  tllws2 = TextListListWithSelection()
  tllws2.tll = tllws.tll[:] # Deep copy not necessary here.
  sel = tllws.selected
  newtl = tllws2.tll[sel][:]
  newtl[0]="Puzzle name: "+stored_param
  tllws2.tll[sel] = newtl
  tllws2.selected = sel
  return tllws2

def replace_starting_file(tllws, stored_param, ps=""):
  #print "In replace_starting_file, stored_param is: "+stored_param
  tllws2 = TextListListWithSelection()
  tllws2.tll = tllws.tll[:] # Deep copy not necessary here.
  sel = tllws.selected
  newtl = tllws2.tll[sel][:]
  newtl[1]="Start with: "+stored_param
  tllws2.tll[sel] = newtl
  tllws2.selected = sel
  return tllws2

def get_puzzle_filename(p): # p is one text list (puzzle).
  line = p[1]
  return line[12:] # Skip "Start with: ".

def replace_embedded_message(tllws, stored_param, ps=""):
  tllws2 = TextListListWithSelection()
  tllws2.tll = tllws.tll[:] # Deep copy not necessary here.
  sel = tllws.selected
  newtl = tllws2.tll[sel][:]
  newtl[2]="Hidden message: "+stored_param
  tllws2.tll[sel] = newtl
  tllws2.selected = sel
  return tllws2



def need_selection(tllws, ps=""):
  return tllws.selected > -1

def reset_highlight(tllws):
  '''If it's obvious that this is a rule, and there is at least one
  condition or at least one action, highlight the last of these.
  '''
  tllws.lineToHighlight = -1
  tll = tllws.tll
  if tllws.selected > -1:  # If no selection, forget it.
    component = tll[tllws.selected]
    if len(component)>3 and component[0]=="IF:": # Must be a rule
      i = len(component)-2 # Start from just before "END."
      if component[i]=="THEN:": i-= 1  # No actions, select last cond.
      tllws.lineToHighlight = i
      
def delete_selected(tllws, ps=""):
  # Delete the currently selected text list.
  tllws2 = TextListListWithSelection()
  s = tllws.selected
  tllws2.tll = tllws.tll[:s-1]+tllws.tll[s+1:] # Deep copy not necessary here.
  tllws2.selected = -1
  reset_highlight(tllws2)
  return tllws2

def select_next(tllws, ps=""):
  tllws2 = TextListListWithSelection()
  #s = tllws.selected
  tllws2.tll = tllws.tll # Deep copy not necessary here.
  n = len(tllws.tll)
  if n==0: tllws2.selected = -1
  elif n==1: tllws2.selected = -1 - tllws.selected
  else: tllws2.selected = (tllws.selected + 1) % n
  reset_highlight(tllws2)
  return tllws2

def select_previous(tllws, ps=""):
  tllws2 = TextListListWithSelection()
  #s = tllws.selected
  tllws2.tll = tllws.tll # Deep copy not necessary here.
  n = len(tllws.tll)
  if n==0: tllws2.selected = -1
  elif n==1: tllws2.selected = -1 - tllws.selected
  else: tllws2.selected = (tllws.selected + n - 1) % n
  reset_highlight(tllws2)
  return tllws2

def need_some_tl(tllws, ps=""):
  return len(tllws.tll) > 0

def make_basic_tll_ops(obj_name="text list",group_name=GROUP_NAME,
                       tll_type="Basic"):
  if tll_type=="Rule":
    op1 = TStar.Operator("Add new "+obj_name, new_rule, group=group_name)
  elif tll_type=="ImageP":
      op1 = TStar.Operator("Add new "+obj_name, new_image_puzzle, group=group_name)
  elif tll_type=="MusicP":
      op1 = TStar.Operator("Add new "+obj_name, new_music_puzzle, group=group_name)
  else:
      op1 = TStar.Operator("Add new "+obj_name, new_tl, group=group_name)
  op3 = TStar.Operator("Delete selected "+obj_name, delete_selected,
                       precondition=need_selection, group=group_name)
  op4 = TStar.Operator("Select next "+obj_name, select_next,
                       precondition=need_some_tl, group=group_name)
  op5 = TStar.Operator("Select previous "+obj_name, select_previous,
                       precondition=need_some_tl, group=group_name)
  return [op1,op3,op4,op5]

def initializer():
    # Returns valid initial data for this kind of state
    tllws = TextListListWithSelection()
    tllws.tll = []
    #[["First text list, line 0", "Line 1"],["Second text list, line 0"]]
    return tllws

def setup():
  global GROUP_NAME, OBJECT_NAME
  GROUP_NAME = "Text List List Designer"
  
  basic_ops = make_basic_tll_ops()

  COStar.register_state_space_component(GROUP_NAME,
                                      TextListListState,
                                      TextListListNode,
                                      basic_ops,
                                      initializer)








